# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Data manipulation steps for standard "single_channel" algorithm.

Generalized algorithm to apply data manipulation steps in a standard order
to apply corrections to a single channel output product.
"""
import logging

LOG = logging.getLogger(__name__)

KtoC_conversion = -273.15

interface = "algorithms"
family = "list_numpy_to_numpy"
name = "single_channel"


def call(
    arrays,
    output_data_range=None,
    input_units=None,
    output_units=None,
    min_outbounds="crop",
    max_outbounds="crop",
    norm=False,
    inverse=False,
    sun_zen_correction=False,
    mask_night=False,
    max_day_zen=None,
    mask_day=False,
    min_night_zen=None,
    gamma_list=None,
    scale_factor=None,
):
    """Apply data range and requested corrections to a single channel product.

    Data manipulation steps for applying a data range and requested corrections
    to a single channel product

    Parameters
    ----------
    arrays : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data
        * MUST be length one for single_channel algorithm.
    output_data_range : list of float, default=None
        * list of min and max value for output data product.
        * This is applied LAST after all other corrections/adjustments
        * If None, use data min and max.
    input_units : str, default=None
        * Units of input data, for applying necessary conversions
        * If None, no conversion
    output_units : str, default=None
        * Units of output data, for applying necessary conversions
        * If None, no conversion
    min_outbounds : str, default='crop'
        * Method to use when applying bounds.  Valid values are:

            * retain: keep all pixels as is
            * mask: mask all pixels that are out of range
            * crop: set all out of range values to either min_val or max_val
              as appropriate
    max_outbounds : str, default='crop'
        * Method to use when applying bounds.  Valid values are:

            * retain: keep all pixels as is
            * mask: mask all pixels that are out of range
            * crop: set all out of range values to either min_val or max_val
              as appropriate
    norm : bool, default=False
        * Boolean flag indicating whether to normalize (True) or not (False)

            * If True, returned data will be in the range from 0 to 1
            * If False, returned data will be in the range from min_val to max_val
    inverse : bool, default=False
        * Boolean flag indicating whether to inverse (True) or not (False)

            * If True, returned data will be inverted
            * If False, returned data will not be inverted
    sun_zenith_correction : bool, default=False
        * Boolean flag indicating whether to apply solar zenith correction
          (True) or not (False)

            * If True, returned data will have solar zenith correction applied
              (see data_manipulations.corrections.apply_solar_zenith_correction)
            * If False, returned data will not be modified based on solar zenith
              angle

    Notes
    -----
    Order of operations, based on the passed arguments, is:

    1. Mask night
    2. Mask day
    3. Apply solar zenith correction
    4. Apply gamma values
    5. Apply scale factor
    6. Convert units
    7. Apply data range.

    NOTE: If "norm=True" is specified, the "output_data_range" will NOT
    match the actual range of the returned data, since the normalized data
    will be returned between 0 and 1.

    If you require a different order of operations than that specified within
    "single_channel" algorithm, please create a new algorithm for your desired
    order of operations.

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of appropriately scaled channel data,
        in units "output_units".
    """
    data = arrays[0]
    if output_data_range is None:
        output_data_range = [data.min(), data.max()]

    # Mask everything greater than max_day_zen
    # day zenith angles are less than 90
    # night zenith angles are greater than 90
    if max_day_zen and len(arrays) == 2:
        from geoips.data_manipulations.info import percent_unmasked
        from geoips.data_manipulations.corrections import mask_night

        sun_zenith = arrays[1]
        LOG.info("Percent unmasked day/night %s", percent_unmasked(data))
        data = mask_night(data, sun_zenith, max_day_zen)
        LOG.info("Percent unmasked day only %s", percent_unmasked(data))

    # Mask everything less than min_night_zen
    # day zenith angles are less than 90
    # night zenith angles are greater than 90
    if mask_day and min_night_zen and len(arrays) == 2:
        from geoips.data_manipulations.info import percent_unmasked
        from geoips.data_manipulations.corrections import mask_day

        sun_zenith = arrays[1]
        LOG.info("Percent unmasked day/night %s", percent_unmasked(data))
        data = mask_day(data, sun_zenith, min_night_zen)
        LOG.info("Percent unmasked night only %s", percent_unmasked(data))

    if sun_zen_correction and len(arrays) == 2:
        sun_zenith = arrays[1]
        from geoips.data_manipulations.corrections import apply_solar_zenith_correction

        data = apply_solar_zenith_correction(data, sun_zenith)

    if gamma_list is not None:
        from geoips.data_manipulations.corrections import apply_gamma

        for gamma in gamma_list:
            data = apply_gamma(data, gamma)

    if scale_factor is not None:
        from geoips.data_manipulations.corrections import apply_scale_factor

        data = apply_scale_factor(data, scale_factor)

    from geoips.data_manipulations.conversions import unit_conversion

    data = unit_conversion(data, input_units, output_units)

    from geoips.data_manipulations.corrections import apply_data_range

    data = apply_data_range(
        data,
        min_val=output_data_range[0],
        max_val=output_data_range[1],
        min_outbounds=min_outbounds,
        max_outbounds=max_outbounds,
        norm=norm,
        inverse=inverse,
    )

    return data



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test all YAML plugins."""
import pytest
import yaml
from importlib import resources, metadata

from geoips.interfaces.base import YamlPluginValidator
from geoips.interfaces.yaml_based.products import ProductsPluginValidator

validator = YamlPluginValidator()
product_validator = ProductsPluginValidator()


def yield_plugins():
    """Yield plugins."""
    plugin_packages = metadata.entry_points(group="geoips.plugin_packages")
    for pkg in plugin_packages:
        fpath = resources.files(pkg.name) / "plugins/yaml"
        plugin_files = fpath.rglob("*.yaml")
        for pf in plugin_files:
            yield pf


def gen_label(val):
    """Generate the yaml name for pytest ids."""
    return val.name


@pytest.mark.parametrize("plugin", yield_plugins(), ids=gen_label)
def test_is_plugin_valid(plugin):
    """Test if plugin is valid."""
    rplugin = yaml.safe_load(open(plugin, "r"))
    if rplugin["interface"] == "products":
        product_validator.validate(rplugin)
    else:
        validator.validate(rplugin)



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Gridline Annotator interface module."""

from geoips.interfaces.base import BaseYamlInterface


class GridlineAnnotatorsInterface(BaseYamlInterface):
    """Interface for gridline annotator plugins."""

    name = "gridline_annotators"


gridline_annotators = GridlineAnnotatorsInterface()



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Matplotlib-based unprojected image output."""

import logging

import matplotlib.pyplot as plt
import matplotlib

from geoips.image_utils.mpl_utils import save_image

matplotlib.use("agg")

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "unprojected"
name = "unprojected_image"


def call(
    xarray_obj,
    product_name,
    output_fnames,
    product_name_title=None,
    mpl_colors_info=None,
    x_size=None,
    y_size=None,
    savefig_kwargs=None,
):
    """Plot unprojected image to matplotlib figure."""
    if savefig_kwargs is None:
        # Default to no arguments, empty dictionary.  Will result in masked background
        savefig_kwargs = {}
        # black background
        # savefig_kwargs = {'facecolor': 'k'}

    if product_name_title is None:
        product_name_title = product_name
    if x_size is None and y_size is None:
        x_size = xarray_obj[product_name].shape[1]
        y_size = xarray_obj[product_name].shape[0]
    elif y_size is None:
        ratio = float(x_size) / float(xarray_obj[product_name].shape[1])
        y_size = float(ratio) * float(xarray_obj[product_name].shape[0])
    elif x_size is None:
        ratio = float(y_size) / float(xarray_obj[product_name].shape[0])
        x_size = float(ratio) * float(xarray_obj[product_name].shape[1])

    rc_params = matplotlib.rcParams
    dpi = rc_params["figure.dpi"]

    image_width = float(x_size) / dpi
    image_height = float(y_size) / dpi

    fig = plt.figure(facecolor="none")
    fig.set_size_inches(image_width, image_height)
    main_ax = plt.Axes(fig, [0, 0, 1, 1])
    main_ax.set_axis_off()
    fig.add_axes(main_ax)

    main_ax.imshow(
        xarray_obj[product_name],
        norm=mpl_colors_info["norm"],
        cmap=mpl_colors_info["cmap"],
    )

    success_outputs = []
    for fname in output_fnames:
        LOG.info("Plotting %s with plt", fname)
        # This just handles cleaning up the axes, creating directories, etc
        success_outputs += save_image(
            fig,
            fname,
            is_final=False,
            image_datetime=xarray_obj.start_datetime,
            savefig_kwargs=savefig_kwargs,
        )

    return success_outputs



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS CLI "config" command.

Various configuration-based commands for setting up your geoips environment.
"""

from numpy import any
from os import listdir, environ
from os.path import abspath, join
import requests
import tarfile

from geoips.commandline.ancillary_info.test_data import test_dataset_dict
from geoips.commandline.geoips_command import GeoipsCommand, GeoipsExecutableCommand


class GeoipsConfigInstall(GeoipsExecutableCommand):
    """Config Command Class for installing packages/data.

    Supports installation of packages and test data needed for testing and/or running
    your GeoIPS environment.
    """

    name = "install"
    command_classes = []

    @property
    def geoips_testdata_dir(self):
        """String path to GEOIPS_TESTDATA_DIR."""
        if not hasattr(self, "_geoips_testdata_dir"):
            self._geoips_testdata_dir = environ["GEOIPS_TESTDATA_DIR"]
        return self._geoips_testdata_dir

    def add_arguments(self):
        """Add arguments to the config-subparser for the Config Command."""
        self.parser.add_argument(
            "test_dataset_name",
            type=str.lower,
            choices=list(test_dataset_dict.keys()),
            help="GeoIPS Test Dataset to Install.",
        )

    def __call__(self, args):
        """Run the `geoips config install <test_dataset_name>` command.

        Parameters
        ----------
        args: Namespace()
            - The argument namespace to parse through
        """
        test_dataset_name = args.test_dataset_name
        test_dataset_url = test_dataset_dict[test_dataset_name]
        if any([test_dataset_name in fol for fol in listdir(self.geoips_testdata_dir)]):
            print(
                f"Test dataset '{test_dataset_name}' already exists under "
                f"'{join(self.geoips_testdata_dir, test_dataset_name)}*/'. See that "
                "location for the contents of the test dataset."
            )
        else:
            print(
                f"Installing {test_dataset_name} test dataset. This may take a while..."
            )
            self.download_extract_test_data(test_dataset_url, self.geoips_testdata_dir)
            out_str = f"Test dataset '{test_dataset_name}' has been installed under "
            out_str += f"{self.geoips_testdata_dir}/{test_dataset_name}/"
            print(out_str)

    def download_extract_test_data(self, url, download_dir):
        """Download the specified URL and write it to the corresponding download_dir.

        Will extract the data using tarfile and create an archive by bundling the
        associated files and directories together.

        Parameters
        ----------
        url: str
            - The url of the test dataset to download
        download_dir: str
            - The directory in which to download and extract the data into
        """
        resp = requests.get(url, stream=True, timeout=15)
        if resp.status_code == 200:
            self.extract_data_cautiously(resp, download_dir)
        else:
            self.parser.error(
                f"Error retrieving data from {url}; Status Code {resp.status_code}."
            )

    def extract_data_cautiously(self, response, download_dir):
        """Extract the GET Response cautiously and skip any dangerous members.

        Iterate through a Response and check that each member is not dangerous to
        extract to your machine. If it is, skip it.

        Where 'dangerous' is a filepath that is not part of 'download_dir'. File path
        maneuvering characters could be invoked ('../', ...), which we will not allow
        when downloading test data.

        Parameters
        ----------
        response: Requests Response Object
            - The GET Response from retrieving the data url
        download_dir: str
            - The directory in which to download and extract the data into
        """
        with tarfile.open(fileobj=response.raw, mode="r|gz") as tar:
            # Validate and extract each member of the archive
            for m in tar:
                if not abspath(join(download_dir, m.name)).startswith(download_dir):
                    raise SystemExit("Found unsafe filepath in tar, exiting now.")
                tar.extract(m, path=download_dir)


class GeoipsConfig(GeoipsCommand):
    """Config top-level command for configuring your GeoIPS environment."""

    name = "config"
    command_classes = [GeoipsConfigInstall]



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Feature Annotator interface module."""

from geoips.interfaces.base import BaseYamlInterface


class FeatureAnnotatorsInterface(BaseYamlInterface):
    """Interface for feature annotator plugins."""

    name = "feature_annotators"


feature_annotators = FeatureAnnotatorsInterface()



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test script for representative product comparisons."""

import subprocess
import logging
from os.path import splitext

LOG = logging.getLogger(__name__)

interface = "output_checkers"
family = "standard"
name = "text"


def clear_text(match_path, close_path, bad_path):
    """Clear output text files so they can be written again."""
    open(match_path, "w").close()
    open(close_path, "w").close()
    open(bad_path, "w").close()


def copy_files(text_path, savedir, file_ext):
    """Copy text file internal in GeoIPS to output directory."""
    from shutil import copy
    from os.path import join

    copy(text_path, join(savedir, "compare." + file_ext))
    copy(text_path, join(savedir, "matched." + file_ext))
    copy(text_path, join(savedir, "close_mismatch." + file_ext))
    copy(text_path, join(savedir, "bad_mismatch." + file_ext))


def get_test_files(test_data_dir):
    """Return a series of varied text files."""
    import numpy as np
    from shutil import copy
    from os import makedirs
    from os.path import exists, join
    from importlib.resources import files

    savedir = join(test_data_dir, "scratch", "unit_tests", "test_text")
    if not exists(savedir):
        makedirs(savedir)
    text_path = str(files("geoips") / "plugins/txt/ascii_palettes/tpw_cimss.txt")
    copy_files(text_path, savedir, "txt")
    comp_path = join(savedir, "compare.txt")
    match_path = join(savedir, "matched.txt")
    close_path = join(savedir, "close_mismatch.txt")
    bad_path = join(savedir, "bad_mismatch.txt")
    clear_text(match_path, close_path, bad_path)
    copy(comp_path, match_path)
    with open(comp_path, mode="r") as comp_txt:
        close_mismatch = open(close_path, "w")
        bad_mismatch = open(bad_path, "w")
        for line in comp_txt.readlines():
            for version in range(2):
                rand = np.random.rand()
                if version == 0:  # Close but mismatched
                    if rand > 0.05:
                        close_mismatch.write(line)
                else:  # Mismatched -- not close
                    if rand > 0.25:
                        bad_mismatch.write(line)
        close_mismatch.close()
        bad_mismatch.close()
    return comp_path, [match_path, close_path, bad_path]


def perform_test_comparisons(plugin, compare_file, test_files):
    """Test the comparison of two text files with the Text Output Checker."""
    for path_idx in range(len(test_files)):
        retval = plugin.module.outputs_match(
            plugin,
            test_files[path_idx],
            compare_file,
        )
        if path_idx == 0:
            assert retval is True
        else:
            assert retval is False


def correct_file_format(fname):
    """Check if fname is a text file.

    Parameters
    ----------
    fname : str
        Name of file to check.

    Returns
    -------
    bool
        True if it is a text file, False otherwise.
    """
    if splitext(fname)[-1] in ["", ".txt", ".text", ".yaml"]:
        with open(fname) as f:
            line = f.readline()
        if isinstance(line, str):
            return True
    return False


def outputs_match(plugin, output_product, compare_product):
    """Check if two text files match.

    Parameters
    ----------
    plugin: OutputCheckerPlugin
        The corresponding geotiff OutputCheckerPlugin that has access to needed methods
    output_product : str
        Full path to current output product
    compare_product : str
        Full path to "good" comparison product

    Returns
    -------
    bool
        Return True if products match, False if they differ
    """
    from geoips.commandline.log_setup import log_with_emphasis

    ret = subprocess.run(
        ["diff", output_product, compare_product],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    LOG.debug(ret.stdout)
    if ret.returncode == 0:
        log_with_emphasis(LOG.info, "GOOD Text files match")
        return True

    out_difftxt = plugin.get_out_diff_fname(compare_product, output_product)
    with open(out_difftxt, "w") as fobj:
        subprocess.call(["diff", output_product, compare_product], stdout=fobj)
    log_with_emphasis(
        LOG.interactive,
        "BAD Text files do NOT match exactly",
        f"output_product: {output_product}",
        f"compare_product: {compare_product}",
        f"out_difftxt: {out_difftxt}",
    )
    return False


def call(plugin, compare_path, output_products):
    """Compare the "correct" text found the list of current output_products.

    Compares files produced in the current processing run with the list of
    "correct" files contained in "compare_path".

    Parameters
    ----------
    plugin: OutputCheckerPlugin
        The corresponding text OutputCheckerPlugin that has access to needed methods
    compare_path : str
        Path to directory of "correct" products - filenames must match output_products
    output_products : list of str
        List of strings of current output products,
        to compare with products in compare_path

    Returns
    -------
    int
        Binary code: 0 if all comparisons were completed successfully.
    """
    retval = plugin.compare_outputs(
        compare_path,
        output_products,
    )
    return retval



#!/bin/env python

# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Download data from a specified URL."""

import sys
import requests
from subprocess import check_output


def download_test_data(url, dest=None):
    """Download the specified URL and write to stdout as bytes.

    Will raise requests.exceptions.HTTPError on failure.
    """
    if ".git" in url:
        print(f"git clone {url} {dest}")
        check_output(["git", "clone", url, dest])
        print("done git clone")
    else:
        resp = requests.get(url, stream=True, timeout=15)
        sys.stdout.buffer.write(resp.raw.read())


if __name__ == "__main__":
    dest = None
    if len(sys.argv) == 3:
        dest = sys.argv[2]
    download_test_data(sys.argv[1], dest)



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS error module."""


class PluginError(Exception):
    """Exception to be raised when there is an error in a GeoIPS plugin."""

    pass


class PluginRegistryError(Exception):
    """Exception to be raised when there is an error in a plugin registry."""

    pass


class CoverageError(Exception):
    """Raise exception on data coverage error."""

    pass


class PluginPackageNotFoundError(Exception):
    """Raise exception stating a certain plugin package was not found."""

    pass


class CliError(Exception):
    """Raise exception on command line interface error."""

    pass



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Geotiff image rasterio-based output format."""

import os
import logging

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "image"
name = "cogeotiff_rgba"


def get_rasterio_cmap_dict(mpl_cmap, scale_data_min=1, scale_data_max=255):
    """Get rasterio cmap dict."""
    cmap_arr = mpl_cmap(range(0, 255)) * 255
    cmap_dict = {}
    for ii in range(0, 255):
        cmap_dict[ii] = tuple(cmap_arr[ii, :])
    return cmap_dict


def scale_geotiff_data(
    plot_data, mpl_colors_info, scale_data_min=1, scale_data_max=255, missing_value=0
):
    """Scale geotiff data."""
    from geoips.data_manipulations.corrections import apply_data_range

    min_val = None
    max_val = None
    inverse = False
    if (
        mpl_colors_info
        and "norm" in mpl_colors_info
        and hasattr(mpl_colors_info["norm"], "vmax")
    ):
        min_val = mpl_colors_info["norm"].vmin
        max_val = mpl_colors_info["norm"].vmax

    num_colors = scale_data_max - scale_data_min  # + 1

    scale_data = (
        scale_data_min
        + apply_data_range(
            plot_data, min_val=min_val, max_val=max_val, inverse=inverse, norm=True
        )
        * num_colors
    )
    scale_data.fill_value = missing_value
    return scale_data.filled()


def call(
    area_def,
    xarray_obj,
    product_name,
    output_fnames,
    product_name_title=None,
    mpl_colors_info=None,
    existing_image=None,
):
    """Create standard geotiff output using rasterio."""
    if area_def.proj_dict["proj"] != "eqc":
        raise TypeError(
            "Projection must be 'eqc' in order to produce COGS output. "
            "This writer relies on equally spaced lat/lons"
        )
    plot_data = scale_geotiff_data(
        xarray_obj[product_name].to_masked_array(), mpl_colors_info
    )
    from rasterio.io import MemoryFile
    from rasterio import uint8  # , uint16
    from rasterio.plot import reshape_as_raster

    # from rasterio.transform import from_bounds
    from rio_cogeo.cogeo import cog_translate
    from rio_cogeo.profiles import cog_profiles

    # https://cogeotiff.github.io/rio-cogeo/API/
    from affine import Affine

    for output_fname in output_fnames:
        from geoips.filenames.base_paths import make_dirs

        make_dirs(os.path.dirname(output_fname))
        with MemoryFile() as memfile:

            # lons, lats = area_def.get_lonlats()

            # Rasterio uses numpy array of shape of `(bands, height, width)`
            height = plot_data.shape[0]
            width = plot_data.shape[1]
            if len(plot_data.shape) == 3:
                bands = plot_data.shape[2]
                plot_data = reshape_as_raster(plot_data.astype(uint8))
                indexes = ()
                for i in range(bands):
                    indexes += (i + 1,)
            else:
                bands = 1
                indexes = 1

            minlat = area_def.area_extent_ll[1]
            maxlat = area_def.area_extent_ll[3]

            minlon = area_def.area_extent_ll[0]
            maxlon = area_def.area_extent_ll[2]
            # Updated for crossing dateline
            if maxlon < 0 and minlon > maxlon:
                maxlon += 360
                raise ValueError(
                    "Region CAN NOT cross the dateline. "
                    "Try again with a different sector"
                )

            res_deg_x = abs(maxlon - minlon) / (width)
            # If we do not add one to height, the calculated min
            # for the lat extends slightly past the range.
            res_deg_y = (maxlat - minlat) / (height + 1)

            # Note the original implementation here, following the
            # rasterio documentation, resulted in upside down imagery.
            # https://github.com/rasterio/rasterio/issues/1683
            # Updates 20231005 resolved the flipped geotiff imagery.
            # Updates 20240322 discovered the rasterio documentation was
            # probably correct, need to flip back.
            # Flipped
            src_transform = Affine.translation(
                minlon - res_deg_x / 2, maxlat - res_deg_y / 2
            ) * Affine.scale(res_deg_x, -res_deg_y)
            # Orig
            # src_transform = Affine.translation(
            #     minlon - res_deg_y / 2, minlat - res_deg_x / 2
            # ) * Affine.scale(res_deg_y, res_deg_x)

            # crs = rasterio.crs.CRS.from_proj4(area_def.proj4_string)
            # Test with "EPSG:4326" crs
            # crs = "EPSG:4326"
            # Flipped
            crs = "+proj=longlat"
            # Orig
            # crs = "+proj=latlong"
            src_profile = dict(
                driver="GTiff",
                dtype=uint8,
                count=bands,
                height=height,
                width=width,
                crs=crs,
                transform=src_transform,
            )

            with memfile.open(**src_profile) as mem:
                # Populate the input file with numpy array
                mem.write(plot_data.astype(uint8), indexes=indexes)

                dst_profile = cog_profiles.get("deflate")
                cog_translate(
                    mem,
                    output_fname,
                    dst_profile,
                    in_memory=True,
                    quiet=True,
                )

    return output_fnames



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Processing workflow for config-based processing."""

import logging
from copy import deepcopy
from glob import glob
from os.path import exists
from os.path import basename
from os import getpid
from datetime import datetime

from geoips.commandline.args import check_command_line_args
from geoips.filenames.base_paths import PATHS as gpaths
from geoips.utils.memusg import PidLog

from geoips.geoips_utils import output_process_times
from geoips.dev.product import (
    get_covg_from_product,
    get_covg_args_from_product,
    get_required_variables,
)
from geoips.xarray_utils.data import (
    sector_xarrays,
    combine_preproc_xarrays_with_alg_xarray,
)
from geoips.filenames.duplicate_files import remove_duplicates
from geoips.geoips_utils import replace_geoips_paths
from geoips.utils.context_managers import import_optional_dependencies

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    from geoips_db.utils.database_writes import (
        write_to_database,
        flag_product_as_deleted,
        write_stats_to_database,
    )

# Old interfaces (YAML, will migrate to new soon)
from geoips.dev.output_config import (
    get_output_formatter_kwargs,
    get_output_formatter,
    get_minimum_coverage,
    produce_current_time,
)

# New interfaces
from geoips.interfaces import interpolators
from geoips.interfaces import output_formatters
from geoips.interfaces import readers
from geoips.interfaces import products
from geoips.interfaces import sector_adjusters

# Collect functions from single_source (should consolidate these somewhere)
from geoips.plugins.modules.procflows.single_source import (
    process_sectored_data_output,
    process_xarray_dict_to_output_format,
    pad_area_definition,
    add_filename_extra_field,
    get_area_defs_from_command_line_args,
    plot_data,
    combine_filename_extra_fields,
    get_alg_xarray,
    verify_area_def,
    remove_unsupported_kwargs,
    get_unique_dataset_key,
)

# Moved to top-level errors module, fixing issue #67
from geoips.errors import CoverageError

PMW_NUM_PIXELS_X = 1400
PMW_NUM_PIXELS_Y = 1400
PMW_PIXEL_SIZE_X = 1000
PMW_PIXEL_SIZE_Y = 1000

LOG = logging.getLogger(__name__)

interface = "procflows"
family = "standard"
name = "config_based"


# get geoips version
try:
    geoips_version = gpaths["GEOIPS_VERS"]
except KeyError:
    LOG.warning("No geoips system defined, setting geoips version to 0.0.0")
    geoips_version = "0.0.0"


def update_output_dict_from_command_line_args(output_dict, command_line_args=None):
    """Update output dict from command line args."""
    if command_line_args is None:
        LOG.info(
            "SKIPPING command_line_args not specified, returning original output_dict"
        )
        return output_dict

    final_output_dict = output_dict.copy()
    for cmdline_fld_name in [
        "filename_formatter_kwargs",
        "metadata_filename_formatter_kwargs",
    ]:
        # Skip fields that are NOT in command_line_args
        if (
            cmdline_fld_name not in command_line_args
            or command_line_args[cmdline_fld_name] is None
        ):
            LOG.info("SKIPPING %s argument not specified command line")
            continue

        # Convert filename_formatter_kwargs and metadata_filename_formatter_kwargs to
        # their plural counterparts
        # Ensure we copy the command_line_args so we do not inadvertently modify
        # the original command line arguments. This should likely be refactored.
        if cmdline_fld_name == "filename_formatter_kwargs":
            output_fld_name = "filename_formatters_kwargs"
            output_fld_val = {"all": deepcopy(command_line_args[cmdline_fld_name])}
        elif cmdline_fld_name == "metadata_filename_formatter_kwargs":
            output_fld_name = "metadata_filename_formatters_kwargs"
            output_fld_val = {"all": deepcopy(command_line_args[cmdline_fld_name])}
        else:
            output_fld_name = cmdline_fld_name
            output_fld_val = deepcopy(command_line_args[cmdline_fld_name])

        # If the current command line field is not in the output dict at all, just
        # add the whole thing.
        if output_fld_name not in output_dict:
            LOG.info(
                'ADDING output_dict "%s" from command_line_args: %s',
                cmdline_fld_name,
                output_fld_val,
            )
            final_output_dict[output_fld_name] = output_fld_val

        # If the special key 'all' is in the desired output_dict field,
        # and it is NOT currently in the
        # output_dict, then add the entire "all" field
        elif "all" in output_fld_val and "all" not in output_dict[output_fld_name]:
            LOG.warning(
                'REPLACING output_dict "%s" *all*with command_line_args: %s',
                output_fld_name,
                output_fld_val,
            )
            final_output_dict[output_fld_name]["all"] = output_fld_val["all"]

        # If the special key "all" is in the desired output_dict field,
        # but IS currently in the output_dict, then add the individual fields
        # within the command line dictionary
        elif "all" in output_fld_val:
            LOG.warning(
                'REPLACING output_dict "%s" *fields* with command_line_args: %s',
                output_fld_name,
                output_fld_val,
            )
            for kwarg_name in output_fld_val["all"]:
                final_output_dict[output_fld_name]["all"][kwarg_name] = output_fld_val[
                    "all"
                ][kwarg_name]

        # Otherwise this is a normal argument, just replace it.
        else:
            LOG.warning(
                'REPLACING entire output_dict "%s" with command_line_args: %s',
                output_fld_name,
                output_fld_val,
            )
            final_output_dict[cmdline_fld_name] = output_fld_val

    return final_output_dict


def get_required_outputs(config_dict, sector_type):
    """Get only the required outputs from the current sector_type."""
    return_dict = {}
    for output_type, output_dict in config_dict["outputs"].items():
        # If the current output type does not require the current sector_type, skip
        if output_dict["requested_sector_type"] != sector_type:
            continue
        return_dict[output_type] = output_dict

    return return_dict


def get_bg_xarray(
    sect_xarrays,
    area_def,
    prod_plugin,
    resampled_read=False,
    window_start_time=None,
    window_end_time=None,
):
    """Get background xarray.

    Parameters
    ----------
    sect_xarrays: dict of xarray.Dataset
        dictionary of xarray Datasets to pull appropriate background xarray from.
        This may include multiple products / variables.
    area_def: pyresample.AreaDefinition
        Spatial region required in the final xarray Datasets.
    prod_plugin: ProductPlugin
        GeoIPS Product Plugin obtained through interfaces.products.get_plugin("name").
    resampled_read: bool, default=False
        Specify whether a resampled read is required, needed for datatypes that
        will be read within "get_alg_xarray"
    window_start_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.
        hours_before_sector_time and hours_after_sector_time are ignored if
        window start/end time are set!
    window_end_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.
        hours_before_sector_time and hours_after_sector_time are ignored if
        window start/end time are set!

    Returns
    -------
    alg_xarray: xarray.Dataset
        xarray Dataset containing the data needed to produce the background for
        overlay imagery.
    """
    interp_plugin = None
    LOG.interactive(
        "Getting background xarray dataset for product '%s'", prod_plugin.name
    )
    if "interpolator" in prod_plugin["spec"]:
        interp_plugin = interpolators.get_plugin(
            prod_plugin["spec"]["interpolator"]["plugin"]["name"]
        )
        interp_args = prod_plugin["spec"]["interpolator"]["plugin"]["arguments"]
        interp_args = remove_unsupported_kwargs(interp_plugin, interp_args)

    alg_xarray = None

    # If this is a preprocessed data file with the final product in it,
    # just pull the final product. Must take out METADATA dataset!
    if (
        len(set(sect_xarrays.keys()).difference({"METADATA"})) == 1
        and prod_plugin.name in list(sect_xarrays.values())[0].variables
    ):
        sect_xarray = list(sect_xarrays.values())[0]

        # Efficiency hit with to_masked_array
        # LOG.info('Min/max %s %s / %s',
        #          product_name,
        #          sect_xarray[product_name].to_masked_array().min(),
        #          sect_xarray[product_name].to_masked_array().max())

        LOG.interactive(
            "Interpolating %s xarray with args %s",
            sect_xarray.source_name,
            str(interp_args),
        )
        alg_xarray = interp_plugin(
            area_def, sect_xarray, alg_xarray, varlist=[prod_plugin.name], **interp_args
        )

        # Efficiency hit with to_masked_array
        # LOG.info('Min/max interp %s %s / %s',
        #          product_name,
        #          alg_xarray[product_name].min(),
        #          alg_xarray[product_name].max())

    # If this is a raw datafile, pull the required variables for applying the given
    # algorithm, and generate the product array.
    else:
        pad_area_def = pad_area_definition(area_def)
        # Ensure pre-processed and raw look the same - this requires applying algorithm
        # to padded sectored data, since that is what is written out to the
        # pre-processed netcdf file, then interpolating to the desired area definition.

        sect_xarray = get_alg_xarray(
            sect_xarrays,
            pad_area_def,
            prod_plugin,
            resampled_read=resampled_read,
            window_start_time=window_start_time,
            window_end_time=window_end_time,
        )
        LOG.interactive(
            "Interpolating %s xarray with varlist %s",
            sect_xarray.source_name,
            prod_plugin.name,
        )
        alg_xarray = interp_plugin(
            area_def, sect_xarray, alg_xarray, varlist=[prod_plugin.name]
        )

    alg_xarray.attrs["registered_dataset"] = True
    alg_xarray.attrs["area_definition"] = area_def
    if prod_plugin.name in alg_xarray.variables:
        alg_xarray = add_filename_extra_field(
            alg_xarray, "background_data", f"bg{prod_plugin.name}"
        )

    return alg_xarray


def get_resampled_read(
    config_dict,
    area_defs,
    area_def_id,
    sector_type,
    reader_plugin,
    reader_kwargs,
    fnames,
    variables,
):
    """Return dictionary of xarray datasets for a given area def.

    Xarrays resampled to area_def
    """
    return get_sectored_read(
        config_dict,
        area_defs,
        area_def_id,
        sector_type,
        reader_plugin,
        reader_kwargs,
        fnames,
        variables,
    )


def get_sectored_read(
    config_dict,
    area_defs,
    area_def_id,
    sector_type,
    reader_plugin,
    reader_kwargs,
    fnames,
    variables,
):
    """Return dictionary of xarray datasets for a given area def.

    Xarrays sectored to area_def
    """
    area_def = area_defs[area_def_id][sector_type]["area_def"]

    if "primary_sector" in config_dict:
        primary_sector_type = area_defs[area_def_id][config_dict["primary_sector"]]
        pad_area_def = primary_sector_type["area_def"]
    else:
        pad_area_def = pad_area_definition(area_def)
    try:
        xobjs = reader_plugin(
            fnames,
            metadata_only=False,
            chans=variables,
            area_def=pad_area_def,
            **reader_kwargs,
        )
    # geostationary satellites fail with IndexError when the area_def does not intersect
    # the data.  Just skip those.  We need a better method for handling this generally,
    # but for now skip IndexErrors.
    except IndexError as resp:
        LOG.error("%s SKIPPING no coverage for %s", resp, area_def)
        return {}
    return xobjs


def get_area_def_list_from_dict(area_defs):
    """Get a list of actual area_defs from full dictionary.

    Dict returned from get_area_defs_from_available_sectors
    """
    list_area_defs = []
    for area_def_id in area_defs:
        for sector_type in area_defs[area_def_id]:
            for ad in area_defs[area_def_id][sector_type]:
                list_area_defs += [area_defs[area_def_id][sector_type]["area_def"]]
    return list_area_defs


def set_comparison_path(output_dict, product_name, output_type, command_line_args=None):
    """Replace variables specified by <varname> in compare_path.

    Parameters
    ----------
    config : dict
        Dictionary of output specifications, containing key "compare_path"
    product_name : str
        Current requested product name, all instances of
        <product> in compare_path replaced with product_name argument
    output_type : str
        Current requested output type, all instances of
        <output> in compare_path replaced with output argument

    Returns
    -------
    str
        Return a single string with the fully specified comparison path for
        current product
    """
    compare_path = None
    if command_line_args is not None and command_line_args["compare_path"] is not None:
        compare_path = command_line_args["compare_path"]

    # If this config has a compare_path specified, replace variables appropriately
    if "compare_path" in output_dict or compare_path is not None:

        if compare_path is None:
            compare_path = output_dict["compare_path"]

        cpath = (
            compare_path.replace("<product>", product_name)
            .replace("<procflow>", "config_based")
            .replace("<output>", output_type)
        )
    # If there is no comparison specified, identify as "no_comparison"
    else:
        cpath = "no_comparison"

    return cpath


def initialize_final_products(final_products, cpath):
    """Initialize the final_products dictionary with cpath dict key if needed.

    Parameters
    ----------
    final_products : dict
        Dictionary of final products, with keys of final required "compare_path"
        Products with no compare_path specified are stored with the key
        "no_comparison"
    cpath : str
        Key to add to final_products dictionary

    Returns
    -------
    dict
        Return final_products dictionary, updated with current "cpath" key:
        final_products[cpath]['files'] = <list_of_files_in_given_cpath>
    """
    if cpath not in final_products:
        final_products[cpath] = {}
        # This is where we store all the files
        final_products[cpath]["files"] = []
        final_products[cpath]["database writes"] = []

    return final_products


def process_unsectored_data_outputs(
    final_products,
    available_outputs_dict,
    available_sectors_dict,
    xobjs,
    variables,
    command_line_args=None,
    write_to_product_db=False,
):
    """Process unsectored data output.

    Loop through all possible outputs, identifying output types that require
    unsectored data output. Produce all required unsectored data output,
    update final_products dictionary accordingly, and
    return final_products dictionary with the new unsectored outputs.

    Parameters
    ----------
    final_products : dict
        Dictionary of final products, with keys of final required "compare_path"
        Products with no compare_path specified are stored with the key
        "no_comparison"
    available_outputs_dict : dict
        Dictionary of all available output product specifications
    available_sectors_dict : dict
        Dictionary of available sector types - we are looking for available
        sectors that contain the "unsectored" keyword.
    xobjs : dict
        Dictionary of xarray datasets, for use in producing unsectored output
        formats
    variables : list
        List of strings of required variables in the given product.

    Returns
    -------
    dict
        Return final_products dictionary, updated with current "cpath" key:
        final_products[cpath]['files'] = <list_of_files_in_given_cpath>
    """
    # These are the different sectors, one for each method of reprojecting or
    # sectoring or resampling the data
    for sector_type in available_sectors_dict:
        # We are looking for a sector_type that has the keyword "unsectored" meaning we
        # want to process the dat before doing anything else to it
        if (
            "unsectored" in available_sectors_dict[sector_type]
            and available_sectors_dict[sector_type]["unsectored"]
        ):
            # Once we've found an "unsectored" data type, we will look for all the
            # output_types in "available_outputs_dict" that use that sector_type
            for output_type in available_outputs_dict:
                output_dict = available_outputs_dict[output_type]
                output_dict = update_output_dict_from_command_line_args(
                    output_dict, command_line_args
                )
                if output_dict["requested_sector_type"] == sector_type:
                    # Now we will produce all of the individual products for the given
                    # output_type/sector_type
                    for product_name in output_dict["product_names"]:
                        # This grabs the compare_path that was requested in the YAML
                        # config, and replaces all instances of <product> with
                        # product_name and all instances of <output> with output_type
                        cpath = set_comparison_path(
                            output_dict, product_name, output_type, command_line_args
                        )
                        # This adds "cpath" to the final_products dictionary, if
                        # necessary
                        final_products = initialize_final_products(
                            final_products, cpath
                        )

                        # This actually produces all the required output files for the
                        # current produsct
                        prod_plugin = products.get_plugin(
                            xobjs["METADATA"].source_name,
                            product_name,
                            output_dict.get("product_spec_override"),
                        )
                        LOG.interactive(
                            "Processing 'xarray_dict_to_output_format' product '%s'",
                            product_name,
                        )
                        out = process_xarray_dict_to_output_format(
                            xobjs, variables, prod_plugin, output_dict
                        )

                        # Add them to the final_products dictionary - comparisons happen
                        # at the end.
                        final_products[cpath]["files"] += out
                        if write_to_product_db:
                            for fprod in out.keys():
                                product_added = write_to_database(
                                    fprod,
                                    product_name,
                                    xobjs,
                                    available_sectors_dict,
                                    output_dict,
                                    geoips_version,
                                )
                                final_products[cpath]["database writes"] += [
                                    product_added
                                ]
    return final_products


def requires_bg(available_outputs_dict, sector_type):
    """Check if current sector requires background imagery.

    Check if a given sector_type is requested for any product_types that also
    require background imagery.

    Parameters
    ----------
    available_outputs_dict : dict
        Dictionary of all requested output_types (specified in YAML config)
    sector_type : str
        sector_type to determine if any output_types that require background
        imagery also request the passed sector_type

    Returns
    -------
    bool
        * True if any output_types that require background imagery require
          the passed "sector_type"
        * False if no output_types require both background imagery and the
          passed "sector_type"
    """
    # Check each output_type in the full config_dict
    for output_type in available_outputs_dict:
        # If the current output_type has an entry for "background_products" that means
        # it requires background imagery If the current output_type also requested the
        # passed "sector_type", then return True.
        if (
            "background_products" in available_outputs_dict[output_type]
            and available_outputs_dict[output_type]["requested_sector_type"]
            == sector_type
        ):
            return True
    # If no output_types required both background_products and the passed
    # "sector_type" then return False
    return False


def is_required_sector_type(available_outputs_dict, sector_type):
    """Check if current sector is required for any outputs.

    Check if a given sector_type is required for any currently requested
    output_types

    Parameters
    ----------
    available_outputs_dict : dict
        Dictionary of all requested output_types (specified in YAML config)
    sector_type : str
        Determine if any output_types require the currently requested
        "sector_type"

    Returns
    -------
    bool
        * True if any output_types require the passed "sector_type"
        * False if no output_types require the passed "sector_type"
    """
    # Go through each output_type currently requested in the YAML config file
    for output_type in available_outputs_dict.keys():
        # If the passed sector_type is requested for any output_type in the YAML
        # config, return True
        if sector_type == available_outputs_dict[output_type]["requested_sector_type"]:
            return True
    # If the passed sector_type is not needed in the YAML config, return False
    return False


def get_config_dict(config_yaml_file):
    """Populate the full config dictionary from a given YAML config file.

    Includes both sector and output specifications.

    Parameters
    ----------
    config_yaml_file : str
        Full path to YAML config file, containing sector and output
        specifications. YAML config files support environment variables in
        entries flagged with !ENV

    Returns
    -------
    dict
        Return dictionary of both sector and output specifications,
        as found in config_yaml_file. The output dictionary references the
        "sector_types" found in the available_sectors dictionary,
        each output_type requests a specific "sector_type" to be used for
        processing.
    """
    # import yaml
    # with open(config_yaml_file, 'r') as f:
    #     config_dict = yaml.safe_load(f)
    # return config_dict
    # This allows environment variables specified by !ENV ${ENVVARNAME}
    from pyaml_env import parse_config

    return parse_config(config_yaml_file)


def get_variables_from_available_outputs_dict(
    available_outputs_dict, source_name, sector_types=None
):
    """Get required variables for all outputs for a given "source_name".

    Outputs specified within the YAML config.

    Parameters
    ----------
    available_outputs_dict : dict
        Dictionary of all requested output_types (specified in YAML config)
    source_name : str
        Find all required variables for the passed "source_name"
    sector_types : list, default=None
        if sector_types list of strings is passed, only include
        output_types that require one of the passed "sector_types"

    Returns
    -------
    list
        List of all required variables for all output products for the given
        source_name
    """
    variables = []
    # Loop through all possible output types
    for output_type in available_outputs_dict:
        # If we requested specific sector_types, only include output_types that
        # require that sector_type
        if (
            sector_types is None
            or available_outputs_dict[output_type]["requested_sector_type"]
            in sector_types
        ):
            # Loop through all products for the given output_type
            for product_name in available_outputs_dict[output_type]["product_names"]:
                # Add all required variables for the current product and source to the
                # list
                prod_plugin = products.get_plugin(
                    source_name,
                    product_name,
                    available_outputs_dict[output_type].get("product_spec_override"),
                )
                variables += get_required_variables(prod_plugin)
    # Return list of all required variables
    return list(set(variables))


def get_area_defs_from_available_sectors(
    available_sectors_dict, command_line_args, xobjs, variables
):
    """Get all required area_defs for the given set of parameters.

    YAML config parameters (config_dict), command_line_args,
    xobjs, and required variables. Command line args override config
    specifications.

    Parameters
    ----------
    available_sectors_dict : dict
        Dictionary of all requested sector_types (specified in YAML config)
    command_line_args : dict
        Dictionary of command line arguments - any command line argument that is
        also a key in available_sectors_dict[<sector_type>] will replace the
        value in the available_sectors_dict[<sector_type>]
    xobjs : dict
        Dictionary of xarray datasets, used in determining start/end time of
        data files for identifying dynamic sectors
    variables : list
        List of required variables, for determining center coverage for TCs

    Returns
    -------
    dict
        Dictionary of required area_defs, with area_def.name as the dictionary
        keys. Based on YAML config-specified available_sectors, and command
        line args

    Notes
    -----
    * Each area_def.name key has one or more "sector_types" associated with it.
    * Each sector_type dictionary contains the actual "requested_sector_dict"
      from the YAML config, and the actual AreaDefinition object that was
      returned.

        * ``area_defs[area_def.name][sector_type]['requested_sector_dict']``
        * ``area_defs[area_def.name][sector_type]['area_def']``
    """
    area_defs = {}

    # Loop through all available sector types
    for sector_type in available_sectors_dict:
        sector_dict = available_sectors_dict[sector_type].copy()

        # If the current sector_type is "unsectored" skip it, because it has no
        # associated sector information
        if "unsectored" in sector_dict and sector_dict["unsectored"]:
            continue

        # command_line_args take priority over config args - if someone passes something
        # in explicitly, it will be used rather than config "default"
        for argname in command_line_args.keys():
            if command_line_args[argname]:
                sector_dict[argname] = command_line_args[argname]

        # Double check if tcdb should be set to false
        if sector_dict.get("trackfiles"):
            sector_dict["tcdb"] = False

        # This is the standard "get_area_defs_from_command_line_args", YAML config
        # specified sector information matches the command line specified sector
        # information
        curr_area_defs = get_area_defs_from_command_line_args(
            sector_dict, xobjs, variables, filter_time=True
        )

        # Loop through the list of area_defs returned by
        # get_area_defs_from_command_line_args, we are going to organize them
        for area_def in curr_area_defs:
            # Use description or name so it includes synoptic time
            # We want each sectorname as a key in the dictionary, with one or more
            # sector_types attached to it. Ie, we may have different sizes/resolutions
            # for the same region, so we want a dictionary of sector_types
            # within the dictionary of area_defs
            if area_def.name not in area_defs:
                # Store the actual sector_dict and area_def in the dictionary
                area_defs[area_def.name] = {
                    sector_type: {
                        "requested_sector_dict": sector_dict,
                        "area_def": area_def,
                    }
                }
            else:
                area_defs[area_def.name][sector_type] = {
                    "requested_sector_dict": sector_dict,
                    "area_def": area_def,
                }
    return area_defs


def call(fnames, command_line_args=None):
    """Workflow for efficiently running all required outputs.

    Includes all sectors and products specified in a YAML output config file.
    Specified via a YAML config file

    Parameters
    ----------
    fnames : list
        List of strings specifying full paths to input file names to process
    command_line_args : dict
        dictionary of command line arguments

    Returns
    -------
    int
        0 for successful completion,
        non-zero for error (incorrect comparison, or failed run)
    """
    ss_pid = getpid()
    pid_track = PidLog(ss_pid, logstr="MEMUSG")

    LOG.interactive("GEOIPS_VERS {}".format(geoips_version))

    process_datetimes = {}
    process_datetimes["overall_start"] = datetime.utcnow()
    final_products = {}
    removed_products = []
    saved_products = []
    num_jobs = 0

    # These args should always be checked
    check_args = [
        "output_config",
        "reader_kwargs",
        "fuse_files",
        "fuse_reader",
        "fuse_reader_kwargs",
        "fuse_resampled_read",
        "fuse_product",
        "filename_formatter_kwargs",
        "metadata_filename_formatter_kwargs",
        "tcdb_sector_list",
        "window_start_time",
        "window_end_time",
        "product_db",
        "product_db_writer_override",
        "output_file_list_fname",
    ]

    check_command_line_args(check_args, command_line_args)
    config_dict = get_config_dict(command_line_args["output_config"])

    if not fnames and "filenames" in config_dict:
        fnames = glob(config_dict["filenames"])

    if not fnames:
        raise IOError(
            "No files found on disk. "
            "Please include valid files either at the command line, "
            "or within the YAML config"
        )
    for fname in fnames:
        if not exists(fname):
            raise IOError(
                f"File '{fname}' not found. "
                "Please include valid files either at the command line, "
                "or within the YAML config"
            )

    output_file_list_fname = command_line_args["output_file_list_fname"]
    reader_kwargs = None
    bg_files = None
    bg_product_name = None
    bg_reader_kwargs = None
    bg_resampled_read = False
    # bg_self_register_dataset = None
    # bg_self_register_source = None

    # 1. get reader_kwargs from command line if specified command line
    # 2. next get reader_kwargs from YAML config_dict if specified within YAML config
    # 3. Finally, if not specified elsewhere, default reader_kwargs to {}
    if command_line_args.get("reader_kwargs"):
        reader_kwargs = command_line_args["reader_kwargs"]
    elif "reader_kwargs" in config_dict:
        reader_kwargs = config_dict.get("reader_kwargs")
    if not reader_kwargs:
        reader_kwargs = {}
    if command_line_args.get("no_presectoring"):
        presector_data = not command_line_args["no_presectoring"]
    elif "no_presectoring" in config_dict:
        presector_data = not bool(config_dict["no_presectoring"])
    else:
        presector_data = True
    if command_line_args.get("output_checker_kwargs") is not None:
        output_checker_kwargs = command_line_args["output_checker_kwargs"]
    else:
        output_checker_kwargs = {}

    # Allow pulling command line arguments from either command line or YAML config.
    # Command line arguments override YAML config

    window_start_time = None
    window_end_time = None
    if command_line_args.get("window_start_time") is not None:
        window_start_time = command_line_args["window_start_time"]
    elif "window_start_time" in config_dict:
        window_start_time = config_dict["window_start_time"]

    if command_line_args.get("window_end_time") is not None:
        window_end_time = command_line_args["window_end_time"]
    elif "window_end_time" in config_dict:
        window_end_time = config_dict["window_end_time"]

    # Allow pulling command line arguments from either command line or YAML config.
    # Command line arguments override YAML config

    window_start_time = None
    window_end_time = None
    if command_line_args.get("window_start_time") is not None:
        window_start_time = command_line_args["window_start_time"]
    elif "window_start_time" in config_dict:
        window_start_time = config_dict["window_start_time"]

    if command_line_args.get("window_end_time") is not None:
        window_end_time = command_line_args["window_end_time"]
    elif "window_end_time" in config_dict:
        window_end_time = config_dict["window_end_time"]

    if command_line_args.get("fuse_files") is not None:
        bg_files = command_line_args["fuse_files"][0]
    elif "fuse_files" in config_dict:
        bg_files = glob(config_dict["fuse_files"])

    if command_line_args.get("fuse_reader") is not None:
        bg_reader_plugin = readers.get_plugin(command_line_args["fuse_reader"][0])
    elif "fuse_reader" in config_dict:
        bg_reader_plugin = readers.get_plugin(config_dict["fuse_reader"])

    # 1. get bg_reader_kwargs from command line if specified command line
    # 2. next get bg_reader_kwargs from YAML config_dict if specified within YAML
    # 3. Finally, if not specified elsewhere, default bg_reader_kwargs to {}
    if command_line_args.get("fuse_reader_kwargs") is not None:
        bg_reader_kwargs = command_line_args["fuse_reader_kwargs"][0]
    elif "fuse_reader_kwargs" in config_dict:
        bg_reader_kwargs = config_dict["fuse_reader_kwargs"]
    if not bg_reader_kwargs:
        bg_reader_kwargs = {}

    if command_line_args.get("fuse_product") is not None:
        bg_product_name = command_line_args["fuse_product"][0]
    elif "fuse_product" in config_dict:
        bg_product_name = config_dict["fuse_product"]

    if command_line_args.get("fuse_resampled_read") is not None:
        bg_resampled_read = command_line_args["fuse_resampled_read"][0]
    elif "fuse_resampled_read" in config_dict:
        bg_resampled_read = config_dict["fuse_resampled_read"]

    # if command_line_args.get("fuse_self_register_dataset") is not None:
    #     bg_self_register_dataset = command_line_args["fuse_self_register_dataset"][0]
    # elif "fuse_self_register_dataset" in config_dict:
    #     bg_self_register_dataset = config_dict["fuse_self_register_dataset"]

    # if command_line_args.get("fuse_self_register_source") is not None:
    #     bg_self_register_source = command_line_args["fuse_self_register_source"][0]
    # elif "fuse_self_register_source" in config_dict:
    #     bg_self_register_source = config_dict["fuse_self_register_source"]

    if command_line_args.get("product_db"):
        product_db = command_line_args["product_db"]
    elif "product_db" in config_dict:
        product_db = config_dict["product_db"]

    else:
        product_db = False

    if command_line_args.get("product_db_writer_override"):
        for sector, database_writer in command_line_args[
            "product_db_writer_override"
        ].items():
            config_dict["available_sectors"][sector] = database_writer

    if command_line_args.get("composite_output_kwargs_override"):
        for sector_output, kwargs in command_line_args[
            "composite_output_kwargs_override"
        ].items():
            config_dict["outputs"][sector_output]["composite_kwargs"] = kwargs

    if bg_files is not None:
        LOG.interactive(
            "Reading background datasets using reader '%s'...", bg_reader_plugin.name
        )
        bg_xobjs = bg_reader_plugin(bg_files, metadata_only=True, **bg_reader_kwargs)
        prod_plugin = products.get_plugin(
            bg_xobjs["METADATA"].source_name,
            bg_product_name,
        )
        bg_variables = get_required_variables(prod_plugin)
        LOG.interactive("Variables '%s'...", bg_variables)

    if product_db:
        from os import getenv

        if not getenv("GEOIPS_DB_URI"):
            raise ValueError("Need to set both $GEOIPS_DB_URI")

    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
    reader_plugin = readers.get_plugin(config_dict["reader_name"])
    LOG.interactive(
        "Reading metadata from datasets using reader '%s'...", reader_plugin.name
    )
    xobjs = reader_plugin(fnames, metadata_only=True, **reader_kwargs)
    source_name = xobjs["METADATA"].source_name

    if not produce_current_time(config_dict, xobjs["METADATA"], output_dict_keys=None):
        LOG.interactive("SKIPPING ALL PROCESSING no products required for current time")
        return 0

    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
    variables = get_variables_from_available_outputs_dict(
        config_dict["outputs"], source_name
    )

    # If this config does not perform a sectored read, just read all the data now
    # Otherwise data will be read within the area_def loop
    sectored_read = False
    resampled_read = False
    if "sectored_read" in config_dict and config_dict["sectored_read"]:
        sectored_read = True
    if "resampled_read" in config_dict and config_dict["resampled_read"]:
        resampled_read = True

    if not resampled_read and not sectored_read:
        pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
        LOG.interactive("Reading full dataset using reader '%s'...", reader_plugin.name)
        xobjs = reader_plugin(
            fnames, metadata_only=False, chans=variables, **reader_kwargs
        )

    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)

    # command_line_args take priority over config args - if someone passes something in
    # explicitly, it will be used rather than config "default"
    area_defs = get_area_defs_from_available_sectors(
        config_dict["available_sectors"], command_line_args, xobjs, variables
    )

    # Check if we have any required unsectored outputs, if so produce here,
    # then continue
    LOG.interactive("\n\n\n\nNEXT Processing any unsectored data outputs...\n\n")
    final_products = process_unsectored_data_outputs(
        final_products,
        config_dict["outputs"],
        config_dict["available_sectors"],
        xobjs,
        variables,
        command_line_args,
        write_to_product_db=product_db,
    )
    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)

    list_area_defs = get_area_def_list_from_dict(area_defs)

    LOG.interactive(
        "\n\n\n\nNEXT Processing %s sectored data outputs...\n\n", len(area_defs)
    )
    area_def_num = 0
    # Loop through each template - register the data once for each template/area_def
    for area_def_id in area_defs:
        area_def_num = area_def_num + 1

        LOG.interactive(
            "\n\n\n\nNEXT area def id: %s (%s of %s)\n\n",
            area_def_id,
            area_def_num,
            len(area_defs),
        )

        bg_alg_xarrays = {}
        # Loop through each sector_type - each sector_type is a different projection /
        # shape / resolution, so we only want to reproject once for each sector_type
        sector_type_num = 0
        for sector_type in area_defs[area_def_id]:
            sector_type_num = sector_type_num + 1

            curr_variables = get_variables_from_available_outputs_dict(
                config_dict["outputs"], source_name, sector_types=[sector_type]
            )

            if not curr_variables:
                LOG.info("No input variables for sector type: %s", sector_type)
                continue

            # If we read separately for each sector (geostationary), then must set
            # xobjs within area_def loop
            if sectored_read:
                pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
                # This will return potentially multiple sectored datasets of different
                # shapes/resolutions. Note currently get_sectored_read and
                # get_resampled_read are identical, because we have no
                # sectored_read based readers.
                LOG.interactive(
                    "Performing sectored read with reader '%s'", reader_plugin.name
                )
                xobjs = get_sectored_read(
                    config_dict,
                    area_defs,
                    area_def_id,
                    sector_type,
                    reader_plugin,
                    reader_kwargs,
                    fnames,
                    curr_variables,
                )
                if not xobjs:
                    continue
            if resampled_read:
                pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
                # This will return one resampled dataset
                # Note currently get_sectored_read and get_resampled_read are identical,
                # because we have no sectored_read based readers.
                LOG.interactive(
                    "Performing resampled read with reader '%s'", reader_plugin.name
                )
                xobjs = get_resampled_read(
                    config_dict,
                    area_defs,
                    area_def_id,
                    sector_type,
                    reader_plugin,
                    reader_kwargs,
                    fnames,
                    curr_variables,
                )
                if not xobjs:
                    continue

            pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
            area_def = area_defs[area_def_id][sector_type]["area_def"]

            # Padded region to ensure we have enough data for recentering, etc.
            # Do NOT pad if we are using a reader_defined or self_register area_def -
            # that indicates we are going to use all of the data we have, so we
            # will not sector
            if area_def.sector_type not in ["reader_defined", "self_register"]:
                pad_area_def = pad_area_definition(
                    area_def, xobjs["METADATA"].source_name
                )
            else:
                pad_area_def = area_def

            pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
            # See if this sector_type is used at all for product output, if not, skip
            # it.
            if not is_required_sector_type(config_dict["outputs"], sector_type):
                LOG.interactive(
                    "\n\n\nSKIPPING sector type: %s, not required for outputs %s",
                    sector_type,
                    config_dict["outputs"].keys(),
                )
                continue
            requested_sector_dict = area_defs[area_def_id][sector_type][
                "requested_sector_dict"
            ]

            LOG.interactive(
                "\n\n\n\nNEXT area def id: %s (%s of %s), "
                "sector_type: %s (%s of %s)\n\n",
                area_def_id,
                area_def_num,
                len(area_defs),
                sector_type,
                sector_type_num,
                len(area_defs[area_def_id]),
            )

            LOG.info("\n\n\n\narea definition: %s", area_def)
            LOG.info("\n\n\n\nrequested sector dict: %s\n\n\n\n", requested_sector_dict)

            # Reduce hours before and after sector time, so we don't get both overpasses
            # from a single. Sector to pad_area_def so we have enough data for
            # recentering.
            process_datetimes[area_def.area_id] = {}
            process_datetimes[area_def.area_id]["start"] = datetime.utcnow()

            # Make sure we grab some around the required data.
            # Do NOT sector if we are using a reader_defined or self_register area_def -
            # that indicates we are going to use all of the data we have, so we will not
            # sector
            if area_def.sector_type not in ["reader_defined", "self_register"]:
                if presector_data:
                    LOG.interactive("Sectoring xarrays, drop=True")
                    # window start/end time override hours before/after sector time.
                    pad_sect_xarrays = sector_xarrays(
                        xobjs,
                        pad_area_def,
                        varlist=curr_variables,
                        hours_before_sector_time=6,
                        hours_after_sector_time=9,
                        drop=True,
                        window_start_time=window_start_time,
                        window_end_time=window_end_time,
                    )
                else:
                    pad_sect_xarrays = xobjs
            else:
                pad_sect_xarrays = xobjs

            pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)

            # See what variables are left after sectoring (could lose some due to
            # day/night)
            all_vars = []
            for key, xobj in pad_sect_xarrays.items():
                # Double check the xarray object actually contains data
                for var in list(xobj.variables.keys()):
                    if xobj[var].count() > 0:
                        all_vars.append(var)

            # If we didn't get any data, continue to the next sector_type
            if len(pad_sect_xarrays) == 0:
                LOG.interactive(
                    "SKIPPING no pad_area_def pad_sect_xarrays returned for %s",
                    area_def.name,
                )
                continue

            # Now we check to see if the current area_def is the closest one to
            # the dynamic time, if appropriate. We could end up with multiple
            # area_defs for a single dynamic sector, and we can't truly test to see
            # how close each one is to the actual data until we sector it...
            # So, check now to see if any of the area_defs in list_area_defs is
            # closer than pad_area_def
            if not verify_area_def(
                list_area_defs,
                pad_area_def,
                pad_sect_xarrays["METADATA"].start_datetime,
                pad_sect_xarrays["METADATA"].end_datetime,
            ):
                LOG.interactive(
                    "SKIPPING duplicate area_def, out of time range, for %s",
                    area_def.name,
                )
                continue

            # Check the config dict to see if this sector_type requests background
            # products
            if bg_files and requires_bg(config_dict["outputs"], sector_type):
                # If we haven't created the bg_alg_xarray for the current sector_type
                # yet, process it and add to the dictionary
                if sector_type not in bg_alg_xarrays:
                    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
                    bg_pad_sect_xarrays = None
                    try:
                        LOG.interactive(
                            "Reading background data with reader '%s'",
                            bg_reader_plugin.name,
                        )
                        bg_xobjs = bg_reader_plugin(
                            bg_files,
                            metadata_only=False,
                            chans=bg_variables,
                            area_def=pad_area_def,
                        )
                        if presector_data:
                            LOG.interactive(
                                "Sectoring background data, variables %s", bg_variables
                            )
                            # window start/end time override hours before/after sector
                            # time.
                            bg_pad_sect_xarrays = sector_xarrays(
                                bg_xobjs,
                                pad_area_def,
                                varlist=bg_variables,
                                hours_before_sector_time=6,
                                hours_after_sector_time=9,
                                drop=True,
                                window_start_time=window_start_time,
                                window_end_time=window_end_time,
                            )
                        else:
                            bg_pad_sect_xarrays = bg_xobjs
                    except CoverageError as resp:
                        LOG.warning(
                            f"{resp} SKIPPING - NO COVERAGE FOR BACKGROUND DATA"
                        )
                    # Only attempt to get bg xarrays if they weren't sectored away to
                    # nothing.
                    bg_prod_plugin = products.get_plugin(
                        bg_xobjs["METADATA"].source_name, bg_product_name
                    )
                    if bg_pad_sect_xarrays:
                        bg_alg_xarrays[sector_type] = get_bg_xarray(
                            bg_pad_sect_xarrays,
                            area_def,
                            bg_prod_plugin,
                            resampled_read=bg_resampled_read,
                        )
            pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)

            # Must adjust the area definition AFTER sectoring xarray (to get valid
            # start/end time
            sector_adjuster = None
            if "sector_adjuster" in config_dict["available_sectors"][sector_type]:
                sector_adjuster = config_dict["available_sectors"][sector_type][
                    "sector_adjuster"
                ]

            adadj_fnames = []
            if sector_adjuster:
                LOG.info("\n\n\n\nAdjusting Area Definition: %s", sector_adjuster)
                LOG.info(
                    "\n\n\n\nBEFORE ADJUSTMENT area definition: %s\n\n\n\n", area_def
                )
                sect_adj_plugin = sector_adjusters.get_plugin(sector_adjuster)
                # Use normal size sectored xarray when running sector_adjuster, not
                # padded. Center time (mintime + (maxtime - mintime)/2) is very slightly
                # different for different size sectored arrays, so for consistency if we
                # change padding amounts, use the fully sectored array for adjusting the
                # area_def.
                if pad_sect_xarrays["METADATA"].source_name not in ["amsu-b", "mhs"]:
                    # The exact sectored arrays, without padding.
                    # Note this must be sectored both before AND after sector_adjuster -
                    # to ensure we both have an accurate center time for adjustments,
                    # and so we get all of the data.
                    if area_def.sector_type not in ["reader_defined", "self_register"]:
                        LOG.interactive(
                            "Sectoring xarrays for sector adjuster '%s'",
                            sector_adjuster,
                        )
                        if presector_data:
                            # window start/end time override hours before/after sector
                            # time.
                            sect_xarrays = sector_xarrays(
                                pad_sect_xarrays,
                                area_def,
                                varlist=curr_variables,
                                hours_before_sector_time=6,
                                hours_after_sector_time=9,
                                drop=True,
                                window_start_time=window_start_time,
                                window_end_time=window_end_time,
                            )
                        else:
                            sect_xarrays = pad_sect_xarrays
                    else:
                        sect_xarrays = pad_sect_xarrays
                    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
                    # If we didn't get any data, continue to the next sector_type
                    # Note we can have coverage for pad_sect_xarrays, but none for
                    # sect_xarrays - ensure we also skip no coverage for sect_xarrays
                    if len(sect_xarrays) == 0:
                        LOG.interactive(
                            "SKIPPING no area_def sect_xarrays returned for %s",
                            area_def.name,
                        )
                        continue
                    if (
                        sect_adj_plugin.family
                        == "list_xarray_list_variables_to_area_def_out_fnames"
                    ):
                        LOG.interactive("Adjusting sector with '%s'", sector_adjuster)
                        area_def, adadj_fnames = sect_adj_plugin(
                            list(sect_xarrays.values()),
                            area_def,
                            curr_variables,
                            config_dict["available_sectors"][sector_type][
                                "adjust_variables"
                            ],
                        )
                    else:
                        LOG.interactive("Adjusting sector with '%s'", sector_adjuster)
                        area_def = sect_adj_plugin(
                            list(sect_xarrays.values()),
                            area_def,
                            curr_variables,
                            config_dict["available_sectors"][sector_type][
                                "adjust_variables"
                            ],
                        )
                else:
                    # AMSU-b specifically needs full swath width... Need a way to
                    # generalize this.
                    if (
                        sect_adj_plugin.family
                        == "list_xarray_list_variables_to_area_def_out_fnames"
                    ):
                        LOG.interactive("Adjusting sector with '%s'", sector_adjuster)
                        area_def, adadj_fnames = sect_adj_plugin(
                            list(pad_sect_xarrays.values()),
                            area_def,
                            curr_variables,
                            config_dict["available_sectors"][sector_type][
                                "adjust_variables"
                            ],
                        )
                    else:
                        LOG.interactive("Adjusting sector with '%s'", sector_adjuster)
                        area_def = sect_adj_plugin(
                            list(pad_sect_xarrays.values()),
                            area_def,
                            curr_variables,
                            config_dict["available_sectors"][sector_type][
                                "adjust_variables"
                            ],
                        )

                cpath = set_comparison_path(
                    config_dict["available_sectors"][sector_type],
                    product_name="archer",
                    output_type="archer",
                    command_line_args=command_line_args,
                )
                final_products = initialize_final_products(final_products, cpath)
                final_products[cpath]["files"] += adadj_fnames

                LOG.info(
                    "\n\n\n\nAFTER ADJUSTMENT area definition: %s\n\n\n\n", area_def
                )

            pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
            # The exact sectored arrays, without padding.
            # Note this must be sectored AFTER sector_adjuster - to ensure we get all
            # the data. Do NOT sector if we are using a reader_defined or self_register
            # area_def - that indicates we are going to use all of the data we have, so
            # we will not sector
            if area_def.sector_type not in ["reader_defined", "self_register"]:
                LOG.interactive(
                    "Sectoring self register xarrays for area_def '%s'", area_def.name
                )
                if presector_data:
                    # window start/end time override hours before/after sector time.
                    sect_xarrays = sector_xarrays(
                        pad_sect_xarrays,
                        area_def,
                        varlist=curr_variables,
                        hours_before_sector_time=6,
                        hours_after_sector_time=9,
                        drop=True,
                        window_start_time=window_start_time,
                        window_end_time=window_end_time,
                    )
                else:
                    sect_xarrays = pad_sect_xarrays
            else:
                sect_xarrays = pad_sect_xarrays

            pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
            # If we didn't get any data, continue to the next sector_type
            # Note we can have coverage for pad_sect_xarrays, but none for sect_xarrays
            # - ensure we also skip no coverage for sect_xarrays
            if len(sect_xarrays) == 0:
                LOG.interactive(
                    "SKIPPING no area_def sect_xarrays returned for %s", area_def.name
                )
                continue

            # Keep track of the applied algorithms in order to prevent redundant
            # algorithm application
            pad_alg_xarrays = {}
            alg_xarrays = {}
            output_num = 0
            required_outputs = get_required_outputs(config_dict, sector_type)
            for output_type, output_dict in required_outputs.items():
                if not produce_current_time(
                    config_dict, xobjs["METADATA"], output_dict_keys=[output_type]
                ):
                    LOG.interactive(
                        """SKIPPING PROCESSING no products required for output_type {0}
                        at current time""".format(
                            output_type,
                        )
                    )
                    continue
                output_dict = update_output_dict_from_command_line_args(
                    output_dict, command_line_args
                )

                output_num = output_num + 1

                LOG.interactive(
                    "\n\n\n\nNEXT area def id: %s (%s of %s), "
                    "sector_type: %s (%s of %s), "
                    "output_type: %s (%s of %s)\n\n",
                    area_def_id,
                    area_def_num,
                    len(area_defs),
                    sector_type,
                    sector_type_num,
                    len(area_defs[area_def_id]),
                    output_type,
                    output_num,
                    len(required_outputs.keys()),
                )

                LOG.info("\n\n\n\narea definition: %s", area_def)

                product_num = 0
                for product_name in output_dict["product_names"]:
                    product_num = product_num + 1
                    prod_plugin = products.get_plugin(
                        pad_sect_xarrays["METADATA"].source_name,
                        product_name,
                        output_dict.get("product_spec_override"),
                    )
                    LOG.info("\n\n\n\nAll area_def_ids: %s", area_defs.keys())
                    LOG.info(
                        "\n\n\n\nAll sector_types: %s", area_defs[area_def_id].keys()
                    )
                    LOG.info(
                        "\n\n\n\nAll output_types: %s", config_dict["outputs"].keys()
                    )
                    LOG.info(
                        "\n\n\n\nAll product_names: %s", output_dict["product_names"]
                    )
                    LOG.info("\n\n\n\nCurrent area definition: %s", area_def)

                    LOG.interactive(
                        "\n\n\n\nNEXT area def id: %s (%s of %s), "
                        "sector_type: %s (%s of %s), "
                        "output_type: %s (%s of %s), "
                        "product_name: %s (%s of %s)\n\n",
                        area_def_id,
                        area_def_num,
                        len(area_defs),
                        sector_type,
                        sector_type_num,
                        len(area_defs[area_def_id]),
                        output_type,
                        output_num,
                        len(required_outputs.keys()),
                        product_name,
                        product_num,
                        len(output_dict["product_names"]),
                    )

                    LOG.info(
                        """\n\n\n\nAll current output_types for sector_type {0}:
                        {1}\n\n\n\n""".format(
                            sector_type,
                            required_outputs.keys(),
                        )
                    )

                    product_variables = get_required_variables(prod_plugin)

                    # Make sure we still have all the required variables after sectoring
                    if not set(product_variables).issubset(all_vars):
                        LOG.interactive(
                            "SKIPPING product %s missing variables %s",
                            product_name,
                            set(product_variables).difference(all_vars),
                        )
                        continue
                    cpath = set_comparison_path(
                        output_dict, product_name, output_type, command_line_args
                    )
                    final_products = initialize_final_products(final_products, cpath)

                    # Produce sectored data output
                    LOG.interactive(
                        "Processing sectored data products for product '%s'",
                        prod_plugin.name,
                    )
                    curr_output_products = process_sectored_data_output(
                        pad_sect_xarrays,
                        product_variables,
                        prod_plugin,
                        output_dict,
                        area_def=area_def,
                    )
                    # If the current product required sectored data processing, skip the
                    # rest of the loop
                    if curr_output_products:
                        final_products[cpath]["files"] += curr_output_products
                        if product_db:
                            for fprod in curr_output_products:
                                LOG.interactive(
                                    "GEOIPS_VERS writing to db {}".format(
                                        geoips_version
                                    )
                                )
                                product_added = write_to_database(
                                    fprod,
                                    product_name,
                                    pad_sect_xarrays["METADATA"],
                                    config_dict["available_sectors"],
                                    output_dict,
                                    geoips_version,
                                    area_def=area_def,
                                )
                                if product_added is not None:
                                    final_products[cpath]["database writes"] += [
                                        product_added
                                    ]
                        continue

                    output_formatter = get_output_formatter(output_dict)
                    output_fmt_plugin = output_formatters.get_plugin(output_formatter)

                    if output_fmt_plugin.family == "xarray_data":
                        # If we're saving out intermediate data file, write out
                        # pad_area_def.
                        if product_name not in pad_alg_xarrays:
                            pad_alg_xarrays[product_name] = get_alg_xarray(
                                pad_sect_xarrays,
                                pad_area_def,
                                prod_plugin,
                                processed_xarrays=pad_alg_xarrays,
                                resector=presector_data,
                                resampled_read=resampled_read,
                                variable_names=product_variables,
                                window_start_time=window_start_time,
                                window_end_time=window_end_time,
                            )
                        else:
                            LOG.info(
                                "  product %s already in pad_alg_xarrays", product_name
                            )
                            LOG.info(
                                "  pad_alg_xarrays datasets: %s",
                                list(pad_alg_xarrays.keys()),
                            )
                            LOG.info(
                                "  pad_alg_xarrays[%s] datasets: %s",
                                product_name,
                                list(pad_alg_xarrays[product_name].variables.keys()),
                            )
                        alg_xarray = pad_alg_xarrays[product_name]
                    elif area_def.sector_type in ["reader_defined", "self_register"]:
                        alg_xarray = get_alg_xarray(
                            pad_sect_xarrays,
                            pad_area_def,
                            prod_plugin,
                            resector=False,
                            resampled_read=resampled_read,
                            variable_names=product_variables,
                            window_start_time=window_start_time,
                            window_end_time=window_end_time,
                        )
                    else:
                        # If we're writing out an image, cut it down to the desired
                        # size.
                        if product_name not in alg_xarrays:
                            alg_xarrays[product_name] = get_alg_xarray(
                                sect_xarrays,
                                area_def,
                                prod_plugin,
                                processed_xarrays=alg_xarrays,
                                resector=presector_data,
                                resampled_read=resampled_read,
                                variable_names=product_variables,
                                window_start_time=window_start_time,
                                window_end_time=window_end_time,
                            )
                            # This is a unique identifier so we can re-use variables
                            # and product arrays appropriately.
                            area_def_key = get_unique_dataset_key(
                                area_def, alg_xarrays[product_name]
                            )
                            alg_xarrays[area_def_key] = alg_xarrays[product_name]
                        else:
                            LOG.info(
                                "  product %s already in alg_xarrays", product_name
                            )
                            LOG.info(
                                "  alg_xarrays datasets: %s", list(alg_xarrays.keys())
                            )
                            LOG.info(
                                "  alg_xarrays[%s] datasets: %s",
                                product_name,
                                list(alg_xarrays[product_name].variables.keys()),
                            )
                        alg_xarray = alg_xarrays[product_name]

                    covg_plugin = get_covg_from_product(
                        prod_plugin,
                        covg_field="image_production_coverage_checker",
                    )
                    covg_args = get_covg_args_from_product(
                        prod_plugin,
                        covg_field="image_production_coverage_checker",
                    )

                    # Set variable_name to prod_plugin.name if not defined.
                    # Always use variable_name if it is defined.
                    # Remove from args so there is not a duplicate are when passing
                    # (since we are passing covg_varname explicitly).
                    # Note get_covg_args_from_product was updated to return a copy of
                    # covg_args, so this does not impact future uses of the product.
                    covg_varname = covg_args.pop("variable_name", prod_plugin.name)
                    # Note variables can be specified as DATASET:VARIABLE,
                    # since this is a preprocessed alg_xarray, and not a
                    # dictionary of datasets, just use the variable name
                    # (we expect the correct variable will exist in this
                    # final processed array)
                    if ":" in covg_varname:
                        covg_varname = covg_varname.split(":")[1]
                    covg = covg_plugin(alg_xarray, covg_varname, area_def, **covg_args)

                    fname_covg_plugin = get_covg_from_product(
                        prod_plugin,
                        covg_field="filename_coverage_checker",
                    )
                    fname_covg_args = get_covg_args_from_product(
                        prod_plugin,
                        covg_field="filename_coverage_checker",
                    )

                    # Set variable_name to prod_plugin.name if not defined.
                    # Always use variable_name if it is defined.
                    # Remove from args so there is not a duplicate are when passing
                    # (since we are passing covg_varname explicitly).
                    # Note get_covg_args_from_product was updated to return a copy of
                    # covg_args, so this does not impact future uses of the product.
                    covg_varname = fname_covg_args.pop(
                        "variable_name", prod_plugin.name
                    )
                    # Note variables can be specified as DATASET:VARIABLE,
                    # since this is a preprocessed alg_xarray, and not a
                    # dictionary of datasets, just use the variable name
                    # (we expect the correct variable will exist in this
                    # final processed array)
                    if ":" in covg_varname:
                        covg_varname = covg_varname.split(":")[1]
                    fname_covg = fname_covg_plugin(
                        alg_xarray, covg_varname, area_def, **fname_covg_args
                    )

                    minimum_coverage = 10
                    config_minimum_coverage = get_minimum_coverage(
                        prod_plugin.name, output_dict
                    )
                    if hasattr(alg_xarray, "minimum_coverage"):
                        minimum_coverage = alg_xarray.minimum_coverage
                    if config_minimum_coverage is not None:
                        minimum_coverage = config_minimum_coverage
                    LOG.interactive(
                        "Required coverage %s for product %s, actual coverage %s "
                        "using %s checker, filename covg %s using %s",
                        minimum_coverage,
                        product_name,
                        covg,
                        covg_plugin.name,
                        fname_covg,
                        fname_covg_plugin.name,
                    )
                    if covg < minimum_coverage and fname_covg < minimum_coverage:
                        LOG.interactive(
                            "Insufficient coverage %s / %s for data products, SKIPPING",
                            covg,
                            fname_covg,
                        )
                        continue
                    composite_kwargs = output_dict.get("composite_kwargs", {})
                    if composite_kwargs.get("composite_products"):
                        from geoips.utils.composite import find_preproc_alg_files

                        # Required kwargs for generating composite
                        comp_settings = output_dict["composite_kwargs"]
                        reader = readers.get_plugin(
                            comp_settings["composite_input_file_reader"]
                        )
                        comp_file_format = comp_settings["composite_input_file_format"]
                        composite_window = comp_settings["composite_window"]

                        # Optional kwargs for generating composite
                        db_query_plugin = comp_settings.get("database_query_module")
                        db_kwargs = config_dict["available_sectors"][sector_type].get(
                            "product_database_writer_kwargs", {}
                        )
                        db_schemas = db_kwargs.get("schema_name")
                        db_tables = db_kwargs.get("table_name")

                        preproc_files = find_preproc_alg_files(
                            product_time=alg_xarray.start_datetime,
                            composite_window=composite_window,
                            sector_name=area_def.area_id,
                            product=product_name,
                            sensor=alg_xarray.source_name,
                            platform=alg_xarray.platform_name,
                            file_format=comp_file_format,
                            product_db=product_db,
                            db_query_plugin=db_query_plugin,
                            db_schemas=db_schemas,
                            db_tables=db_tables,
                        )
                        if preproc_files:
                            pre_proc = reader(preproc_files)
                            if (
                                "rgb"
                                in prod_plugin["spec"]["colormapper"]["plugin"]["name"]
                            ):
                                rgb_var = prod_plugin.name
                            else:
                                rgb_var = None
                            alg_xarray = combine_preproc_xarrays_with_alg_xarray(
                                pre_proc, alg_xarray, rgb_var=rgb_var
                            )
                            comp_covg = covg_plugin(
                                alg_xarray, covg_varname, area_def, **covg_args
                            )
                            LOG.info("Composite coverage: %s", comp_covg)
                        else:
                            LOG.info("No files to create composite!")

                    plot_data_kwargs = get_output_formatter_kwargs(
                        output_dict,
                        alg_xarray,
                        area_def,
                        sector_type,
                        bg_files,
                        bg_alg_xarrays,
                        bg_product_name,
                    )

                    if (
                        bg_files
                        and "background_products" in output_dict
                        and sector_type in bg_alg_xarrays
                    ):
                        alg_xarray = combine_filename_extra_fields(
                            bg_alg_xarrays[sector_type], alg_xarray
                        )

                    curr_products = plot_data(
                        output_dict,
                        alg_xarray,
                        area_def,
                        prod_plugin,
                        plot_data_kwargs,
                    )

                    if curr_products:
                        final_products[cpath]["files"] += list(curr_products.keys())
                    else:
                        continue

                    if product_db:
                        for fprod in curr_products.keys():
                            product_added = write_to_database(
                                fprod,
                                product_name,
                                alg_xarray,
                                config_dict["available_sectors"],
                                output_dict,
                                geoips_version,
                                coverage=covg,
                                area_def=area_def,
                            )
                            if product_added is not None:
                                final_products[cpath]["database writes"] += [
                                    product_added
                                ]

                    if (
                        "remove_duplicates" in output_dict
                        and output_dict["remove_duplicates"] is not None
                    ):
                        curr_removed_products, curr_saved_products = remove_duplicates(
                            curr_products, remove_files=True
                        )
                        removed_products += curr_removed_products
                        saved_products += curr_saved_products

                    process_datetimes[area_def.area_id]["end"] = datetime.utcnow()
                    num_jobs += 1

    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
    process_datetimes["overall_end"] = datetime.utcnow()

    LOG.interactive("\n\n\nProcessing complete! Checking outputs...\n\n\n")

    retval = 0
    failed_compares = {}
    for cpath in final_products:
        if cpath != "no_comparison":
            from geoips.interfaces.module_based.output_checkers import output_checkers

            for output_product in final_products[cpath]["files"]:
                plugin_name = output_checkers.identify_checker(output_product)
                output_checker = output_checkers.get_plugin(plugin_name)
                kwargs = {}
                if output_checker.name in output_checker_kwargs:
                    kwargs = output_checker_kwargs[output_checker.name]
                curr_retval = output_checker(
                    output_checker,
                    cpath,
                    [output_product],
                    **kwargs,
                )
                retval += curr_retval
                if curr_retval != 0:
                    failed_compares[cpath] = curr_retval
        else:
            LOG.info("No comparison specified, not attempting to compare outputs")

    successful_comparison_dirs = 0
    failed_comparison_dirs = 0

    LOG.interactive(
        "\n\n\nThe following products were produced from procflow %s\n\n",
        basename(__file__),
    )
    for cpath in final_products:
        if cpath in failed_compares:
            LOG.interactive(
                "%s FAILED COMPARISONS IN DIR: %s\n", failed_compares[cpath], cpath
            )
            failed_comparison_dirs = failed_comparison_dirs + 1
        elif cpath != "no_comparison":
            LOG.info("SUCCESSFUL COMPARISON DIR: %s\n", cpath)
            successful_comparison_dirs = successful_comparison_dirs + 1
        for filename in final_products[cpath]["files"]:
            LOG.interactive(
                "    \u001b[34mCONFIGSUCCESS\033[0m %s",
                replace_geoips_paths(filename, curly_braces=True),
            )
            if filename in final_products[cpath]["database writes"]:
                LOG.interactive("    DATABASESUCCESS %s", filename)
        LOG.info("\n")

    for removed_product in removed_products:
        LOG.interactive("    DELETEDPRODUCT %s", removed_product)
        if product_db:
            LOG.info("    FLAGGING as deleted in product database")
            flag_product_as_deleted(removed_product, area_defs)

    if output_file_list_fname:
        LOG.info("Writing successful outputs to %s", output_file_list_fname)
        with open(output_file_list_fname, "w", encoding="utf8") as fobj:
            for cpath in final_products:
                LOG.info("Trying %s", cpath)
                if len(final_products[cpath]["files"]) > 0:
                    LOG.info(
                        "  WRITING %s to output file list, %s products generated",
                        cpath,
                        len(final_products[cpath]["files"]),
                    )
                    fobj.writelines(
                        "\n".join(
                            [
                                fname.replace(
                                    gpaths["GEOIPS_OUTDIRS"], "$GEOIPS_OUTDIRS"
                                )
                                for fname in final_products[cpath]["files"]
                            ]
                        )
                    )
                    # If we don't write out the last newline, then wc won't return the
                    # appropriate number, and we won't get to the last file when
                    # attempting to loop through
                    fobj.writelines(["\n"])
                else:
                    LOG.info(
                        """  SKIPPING WRITING {0} to output file list, no products
                        generated""".format(
                            cpath,
                        )
                    )

    mem_usage_stats = pid_track.print_mem_usg(logstr="MEMUSG", verbose=True)
    LOG.interactive("READER_NAME: %s", config_dict["reader_name"])
    num_products = sum(
        [len(final_products[cpath]["files"]) for cpath in final_products]
    )
    LOG.interactive("NUM_PRODUCTS: %s", num_products)
    if product_db:
        LOG.info(
            "NUM_DATABASE_WRITES: %s",
            sum(
                [
                    len(final_products[cpath]["database writes"])
                    for cpath in final_products
                ]
            ),
        )
    LOG.interactive("NUM_DELETED_PRODUCTS: %s", len(removed_products))
    LOG.interactive("NUM_SUCCESSFUL_COMPARISON_DIRS: %s", successful_comparison_dirs)
    LOG.interactive("NUM_FAILED_COMPARISON_DIRS: %s", failed_comparison_dirs)
    output_process_times(process_datetimes, num_jobs)
    LOG.interactive("GEOIPS_VERS {}".format(geoips_version))
    if product_db:
        all_sectors_use_tcdb = all(
            [
                config_dict["available_sectors"][x].get("tcdb")
                for x in config_dict["available_sectors"].keys()
            ]
        )
        if (
            command_line_args.get("tcdb")
            or command_line_args.get("trackfiles")
            or all_sectors_use_tcdb
        ):
            sector_type = "dynamic_tc"
        else:
            sector_type = "static"

        write_stats_to_database(
            procflow_name="config_based",
            platform=xobjs["METADATA"].platform_name.lower(),
            geoips_vers=geoips_version,
            source=xobjs["METADATA"].source_name,
            product="multi",
            sector_type=sector_type,
            process_times=process_datetimes,
            num_products_created=num_products,
            num_products_deleted=len(removed_products),
            resource_usage_dict=mem_usage_stats,
        )
    else:
        LOG.interactive("NO PRODDB GEOIPS_VERS {}".format(geoips_version))
    return retval



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Code to implement GeoIPS Command Line Interface (CLI).

Will implement a plethora of commands, but for the meantime, we'll work on
'geoips list' and 'geoips run'
"""

from os.path import basename
import sys

from colorama import Fore, Style

from geoips.commandline.cmd_instructions import get_instructions
from geoips.commandline.geoips_command import GeoipsCommand
from geoips.commandline.geoips_config import GeoipsConfig
from geoips.commandline.geoips_describe import GeoipsDescribe
from geoips.commandline.geoips_list import GeoipsList
from geoips.commandline.geoips_run import GeoipsRun
from geoips.commandline.geoips_test import GeoipsTest
from geoips.commandline.geoips_tree import GeoipsTree
from geoips.commandline.geoips_validate import GeoipsValidate


# Logging utitilies will be set up once PR 659, Apply Scope to CLI Arguments, has been
# merged. Until that point we'll set up no logging in the CLI as it will duplicate all
# log statements used in our procflows.


class GeoipsCLI(GeoipsCommand):
    """Top-Level Class for the GeoIPS Commandline Interface (CLI).

    This class includes a list of Command Classes, which will implement the core
    functionality of the CLI. This includes the following as of right now:
    - [GeoipsConfig, GeoipsGet, GeoipsList, GeoipsRun, GeoipsTest, GeoipsValidate]
    """

    name = "geoips"  # Needed since we inherit from GeoipsCommand
    command_classes = [
        GeoipsConfig,
        GeoipsDescribe,
        GeoipsList,
        GeoipsRun,
        GeoipsTest,
        GeoipsTree,
        GeoipsValidate,
    ]

    def __init__(self, instructions_dir=None, legacy=False):
        """Initialize the GeoipsCLI and each of it's command classes.

        The CLI contains a single top-level argparse.ArgumentParser() which contains
        subparsers related to each subcommand. This ensures that each command has a
        unique set of arguments inheirted from command -> subcommand -> subcommand,
        and so on. For example, the GeoipsList Command Class's arguments are inherited
        by all subcommand child classes, which recursively can have their own child
        subcommand classes.

        Parameters
        ----------
        instructions_dir: str or Posix.Path
            - The path to the directory which includes the commandline instructions.
              This is only used for testing purposes so we can ensure the correct
              functionality occurs for possibly missing / invalid instruction files.
        """
        if instructions_dir:
            # Instructions dir has been provided, use the instructions found in that
            # directory so we can test that the correct functionality occurs for any
            # given instruction file state.
            self.cmd_instructions = get_instructions(instructions_dir)
        else:
            # Otherwise use the default instructions which we know are correct
            # (and if they're not, the appropriate error will be raised.)
            self.cmd_instructions = None
        super().__init__(legacy=legacy)

    def execute_command(self):
        """Execute the given command."""
        self.GEOIPS_ARGS = self.parser.parse_args()
        # NOTE: We should discuss how we want to share the selected LOG to child classes
        # They can all use the functionality below, however that would be redundant and
        # there is likely a better way to fix this. Unfortunately 'parse_known_args'
        # didn't work and this is our best solution for the time being.

        # self.LOG = getattr(LOG, self.GEOIPS_ARGS.log_level)
        # self.LOG("LOG LEVEL = {self.GEOIPS_ARGS.log_level}")
        if hasattr(self.GEOIPS_ARGS, "exe_command"):
            # The command called is executable (child of GeoipsExecutableCommand)
            # so execute that command now.
            self.GEOIPS_ARGS.exe_command(self.GEOIPS_ARGS)
        else:
            print(
                # f'"{self.GEOIPS_ARGS.command}" command requires a subcommand.\n\n'
                f"{self.GEOIPS_ARGS.command_parser.format_usage()}"
            )
            sys.exit(2)


def support_legacy_procflows():
    """Run a series of checks on sys.argv to support legacy calls to run_procflow.

    The new GeoIPS CLI has changed how we run our process workflows from now on. Instead
    of calling 'run_procflow' or 'data_fusion_procflow', we now call
    'geoips run single_source' or 'geoips run data_fusion'. This function parses through
    sys.argv and performs the necessary translations to match our current format so that
    the CLI's argparser can call the appropriate functionality.

    Returns
    -------
    legacy: bool
        - The truth value as to whether or not a legacy procflow call was used
    """
    defined_procflow = None
    supported_procflows = ["config_based", "data_fusion", "single_source"]
    if (
        basename(sys.argv[0]) == "geoips"
        and len(sys.argv) > 2
        and sys.argv[1] == "run"
        and (len(sys.argv) < 3 or sys.argv[2] not in supported_procflows)
    ):
        # Either a procflow was not specified or it was an invalid procflow. Notify the
        # user of that with a 'NotImplementedError'.
        if len(sys.argv) >= 3:
            procflow_name = sys.argv[2]
        else:
            procflow_name = "No procflow supplied."
        raise NotImplementedError(
            f"'geoips run' was called alongside procflow: '{procflow_name}'.\nIf you "
            "did not supply a procflow name, this is not supported currently.\n"
            "Eventually, 'geoips run' will call the 'order_based' procflow, however "
            "this is not at the current time of use.\nFor a list of supported "
            f"procflows, choose one of the following: {supported_procflows}."
        )
    elif basename(sys.argv[0]) == "run_procflow":
        entrypoint = "run_procflow"
        defined_procflow = "single_source"

    elif basename(sys.argv[0]) == "data_fusion_procflow":
        entrypoint = "data_fusion_procflow"
        defined_procflow = "data_fusion"

    if defined_procflow:
        # Either 'run_procflow' or 'data_fusion_procflow' was called, make the
        # appropriate translations in sys.argv so the CLI's parser can call the correct
        # functionality
        if "--procflow" in sys.argv:
            # If '--procflow' was found in the command line arguments, loop through the
            # arguments and grab the correct procflow.
            for idx, arg in enumerate(sys.argv):
                if arg == "--procflow" and len(sys.argv) > idx + 1:
                    defined_procflow = sys.argv[idx + 1]
        sys.argv[0] = sys.argv[0].replace(entrypoint, "geoips")
        sys.argv.insert(1, "run")
        sys.argv.insert(2, defined_procflow)
        return True
    return False


def print_beta_warning():
    """Notify the user that the CLI is still in Beta development stage."""
    print(
        Fore.RED
        + "\nWARNING: "
        + Fore.YELLOW
        + "The GeoIPS CLI is currently under development and is subject "
        "to change.\nUntil this warning is removed, do not rely on the CLI to be "
        "static.\nPlease feel free to test the CLI and report any bugs or comments as "
        "an issue here:\n"
        + Fore.BLUE
        + "https://github.com/NRLMMD-GEOIPS/geoips/issues/new/choose\n"
        + Style.RESET_ALL
    )


def main():
    """Entry point for GeoIPS command line interface (CLI)."""
    legacy = support_legacy_procflows()
    # Initialize the CLI and all of its commands
    geoips_cli = GeoipsCLI(legacy=legacy)
    # Execute the called command
    geoips_cli.execute_command()
    # Notify that the user is in Beta development status right now.
    print_beta_warning()


if __name__ == "__main__":
    main()



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Ancillary module containing test dataset information."""

"""Dictionary mapping of GeoIPS Test Datasets.

Mapping goes {"test_dataset_name": "test_dataset_url"}
"""

interface = None  # denotes that this is not a plugin module

test_dataset_dict = {
    "test_data_fusion": r"https://io2.cira.colostate.edu/s/J73tEcn22smktMi/download?path=%2F&files=test_data_fusion.tgz",  # NOQA
    "test_data_noaa_aws": r"https://io2.cira.colostate.edu/s/J73tEcn22smktMi/download?path=%2F&files=test_data_noaa_aws.tgz",  # NOQA
    "test_data_amsr2": r"https://io2.cira.colostate.edu/s/J73tEcn22smktMi/download?path=%2F&files=test_data_amsr2_1.6.0.tgz",  # NOQA
    "test_data_clavrx": r"https://io2.cira.colostate.edu/s/J73tEcn22smktMi/download?path=%2F&files=test_data_clavrx_1.10.0.tgz",  # NOQA
    "test_data_gpm": r"https://io2.cira.colostate.edu/s/J73tEcn22smktMi/download?path=%2F&files=test_data_gpm_1.6.0.tgz",  # NOQA
    "test_data_sar": r"https://io2.cira.colostate.edu/s/J73tEcn22smktMi/download?path=%2F&files=test_data_sar_1.12.2.tgz",  # NOQA
    "test_data_scat": r"https://io2.cira.colostate.edu/s/J73tEcn22smktMi/download?path=%2F&files=test_data_scat_1.11.3.tgz",  # NOQA
    "test_data_smap": r"https://io2.cira.colostate.edu/s/J73tEcn22smktMi/download?path=%2F&files=test_data_smap_1.6.0.tgz",  # NOQA
    "test_data_viirs": r"https://io2.cira.colostate.edu/s/J73tEcn22smktMi/download?path=%2F&files=test_data_viirs_1.6.0.tgz",  # NOQA
}



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS PKGNAME documentation build configuration file."""
import sys
import os
import jinja2
import geoips

VERSION = str(geoips.__version__)

source_path = os.path.dirname(os.path.abspath(__file__))

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.append(os.path.abspath('.'))
sys.path.insert(0, os.path.abspath("../sphinxext"))
sys.path.extend(
    [
        # numpy standard doc extensions
        os.path.join(os.path.dirname(__file__), "..", "../..", "sphinxext")
    ]
)

# Use Jinja for navigational tools
with open(os.path.join(source_path, "_templates/indexrst.html")) as f:
    t = jinja2.Template(f.read())
with open(os.path.join(source_path, "index.rst"), "w") as f:
    f.write(t.render())

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = "5.3.0"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_design",
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinxarg.ext",
    "sphinxcontrib.autoprogram",
    "m2r2",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = [".rst"]

# The encoding of source files.
source_encoding = "utf-8"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "GeoIPS, Geolocated Information Processing System"
copyright = "NOT APPLICAPABLE"
author = "the U.S. NAVAL RESEARCH LABORATORY"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents. Consider setting version=release=VERSION
#
# The full version, including alpha/beta/rc tags.
release = "{0}".format(VERSION)
# The short X.Y version.
version = "{0}".format(VERSION[0:3])

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

exclude_patterns = [
    "**.ipynb_checkpoints",
    # to ensure that include files (partial pages) aren't built, exclude them
    # https://github.com/sphinx-doc/sphinx/issues/1965#issuecomment-124732907
    "**/includes/**",
]

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "pydata_sphinx_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "external_links": [],
    "navigation_with_keys": False,
    "footer_end": ["geoips_footer"],
    "github_url": "https://github.com/NRLMMD-GEOIPS/PKGNAME",
    "navbar_end": ["theme-switcher", "navbar-icon-links"],
    "logo": {
        "image_light": "_static/NRL_logo_RGB.jpg",
        "image_dark": "_static/NRL_logo_sidebar_Reverse.png",
    },
}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/NRL_logo_sidebar_Reverse.png"

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or
# 32x32 pixels large.
html_favicon = "_static/nrlicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%b %d, %Y %H:%M:%S"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# US Navy does not copyright
html_show_copyright = False

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr'
html_search_language = "en"

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
# html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
# html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = "GeoIPS_PKGNAME_doc"

# -- Options for LaTeX output ---------------------------------------------

# Latex additional files to include to format header footer for PDF.
latex_additional_files = ["fancyhf.sty"]

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    "pointsize": "10pt",
    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
    "preamble": r"\usepackage{fancyhf}\usepackage{enumitem}\setlistdepth{99}",
    # Latex figure (float) alignment
    "figure_align": "H",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
pkgnamelatex = "PKGNAME".split("_")
if len(pkgnamelatex) > 1:
    # set up latex escape
    pkgnamelatex = r"\_".join(pkgnamelatex)
latex_documents = [
    (
        master_doc,
        "GeoIPS_PKGNAME.tex",
        "GeoIPS {0} Documentation".format(pkgnamelatex),
        "U.S. NAVAL RESEARCH LABORATORY",
        "manual",
        False,
    ),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = "_static/NRL_logo_RGB.jpg"

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, "GeoIPS_PKGNAME", "GeoIPS PKGNAME Documentation", [author], 1)
]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "GeoIPS_PKGNAME",
        "GeoIPS PKGNAME Documentation",
        author,
        "NRL-MMD",
        "Geolocated Information Processing System.",
        "packages: interface, modules, tools",
    ),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = False


# setup and rstjinja based on pandas-doc rendering
def rstjinja(app, docname, source):
    """Render our pages as a jinja template for fancy templating goodness."""
    # https://www.ericholscher.com/blog/2016/jul/25/integrating-jinja-rst-sphinx/
    # Make sure we're outputting HTML
    if app.builder.format != "html":
        return
    src = source[0]
    rendered = app.builder.templates.render_string(src, app.config.html_context)
    source[0] = rendered


# need min:max width limits ref https://stackoverflow.com/a/43186995/1480918
def setup(app):
    """Css width setup."""
    app.add_css_file("set_width.css")
    """Jinja setup function."""
    app.connect("source-read", rstjinja)



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Geotiff image rasterio-based output format."""

import os
import logging

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "image"
name = "geotiff_standard"


def get_rasterio_cmap_dict(mpl_cmap, scale_data_min=1, scale_data_max=255):
    """Get rasterio cmap dict."""
    cmap_arr = mpl_cmap(range(0, 255)) * 255
    cmap_dict = {}
    for ii in range(0, 255):
        cmap_dict[ii] = tuple(cmap_arr[ii, :])
    return cmap_dict


def scale_geotiff_data(
    plot_data, mpl_colors_info, scale_data_min=1, scale_data_max=255, missing_value=0
):
    """Scale geotiff data."""
    from geoips.data_manipulations.corrections import apply_data_range

    min_val = None
    max_val = None
    inverse = False
    if (
        mpl_colors_info
        and "norm" in mpl_colors_info
        and hasattr(mpl_colors_info["norm"], "vmax")
    ):
        min_val = mpl_colors_info["norm"].vmin
        max_val = mpl_colors_info["norm"].vmax

    num_colors = scale_data_max - scale_data_min + 1

    scale_data = (
        scale_data_min
        + apply_data_range(
            plot_data, min_val=min_val, max_val=max_val, inverse=inverse, norm=True
        )
        * num_colors
    )
    scale_data.fill_value = missing_value
    return scale_data.filled()


def call(
    area_def,
    xarray_obj,
    product_name,
    output_fnames,
    product_name_title=None,
    mpl_colors_info=None,
    existing_image=None,
):
    """Create standard geotiff output using rasterio."""
    plot_data = scale_geotiff_data(
        xarray_obj[product_name].to_masked_array(), mpl_colors_info
    )
    import rasterio
    from affine import Affine

    for output_fname in output_fnames:
        from geoips.filenames.base_paths import make_dirs

        make_dirs(os.path.dirname(output_fname))
        with rasterio.Env():
            lons, lats = area_def.get_lonlats()

            height = plot_data.shape[0]
            width = plot_data.shape[1]

            res_deg_x = (lons[-1][-1] - lons[0][0]) / width
            res_deg_y = (lats[0][0] - lats[-1][-1]) / height

            # Flipped
            maxlat = area_def.area_extent_ll[-1]
            # Orig
            # minlat = area_def.area_extent_ll[1]
            minlon = area_def.area_extent_ll[0]

            # Note the original implementation here, following the
            # rasterio documentation, resulted in upside down imagery.
            # https://github.com/rasterio/rasterio/issues/1683
            # Updates 20231005 resolved the flipped geotiff imagery.
            # Flipped
            transform = Affine.translation(
                minlon - res_deg_x / 2, maxlat - res_deg_y / 2
            ) * Affine.scale(res_deg_x, -res_deg_y)
            # Orig
            # transform = Affine.translation(
            #     minlon - res_deg_y / 2, minlat - res_deg_x / 2
            # ) * Affine.scale(res_deg_y, res_deg_x)

            # crs = rasterio.crs.CRS.from_proj4(area_def.proj4_string)
            # Test with "EPSG:4326" crs
            # crs = "EPSG:4326"
            # Flipped
            crs = "+proj=longlat"
            # Orig
            # crs = "+proj=latlong"

            profile = rasterio.profiles.DefaultGTiffProfile(count=1)
            profile.update(dtype=rasterio.uint8, count=1, compress="lzw")

            with rasterio.open(
                output_fname,
                "w",
                width=width,
                height=height,
                crs=crs,
                transform=transform,
                **profile,
            ) as dst:
                cmap_dict = get_rasterio_cmap_dict(mpl_colors_info["cmap"])
                dst.write_colormap(1, cmap_dict)
                dst.write(plot_data.astype(rasterio.uint8), indexes=1)

    return output_fnames



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read AMSU-B and MHS MIRS NetCDF data files.

This reader is desgined for importing the Advanced Microwave Sounding Unit
(AMSU)-B/Microwave Humidity Sounder (HMS) and EUMETSAQT MHS data files in
hdf5 from NOAA MIRS. These new data files have a different file name convention
and and data structure from previous MSPPS.  Example of MIRIS files:

* NPR-MIRS-IMG_v11r4_ma1_s202101111916000_e202101112012000_c202101112047200.nc

    * (ma1 is for metop-B)
* NPR-MIRS-IMG_v11r4_ma2_s202101111715000_e202101111857000_c202101111941370.nc

    * (ma2 is for metop-A)
* NPR-MIRS-IMG_v11r4_n19_s202101111730000_e202101111916000_c202101112001590.nc

    * (NOAA-19)

AMSU-A channel information::

    Chan# / Freq(GHz) / bands / Bandwidth(MHz) / Beamwidth(deg) / NE#T(K) /
                               (Spec.) Polarization at nadir / Instrument Component
    1	23.800	1	270	3.3	0.30	V	A2
    2	31.400	1	180	3.3	0.30	V	A2
    3	50.300	1	180	3.3	0.40	V	A1-2
    4	52.800	1	400	3.3	0.25	V	A1-2
    5	53.596 +-115	2	170	3.3	0.25	H	A1-2
    6	54.400	1	400	3.3	0.25	H	A1-1
    7	54.940	1	400	3.3	0.25	V	A1-1
    8	55.500	1	330	3.3	0.25	H	A1-2
    9	f0=57,290.344	1	330	3.3	0.25	H	A1-1
    10	f0+-217	2	78	3.3	0.40	H	A1-1
    11	f0+-322.2+-48	4	36	3.3	0.40	H	A1-1
    12	f0+-322.2+-22	4	16	3.3	0.60	H	A1-1
    13	f0+-322.2+-10	4	8	3.3	0.80	H	A1-1
    14	f0+-322.2+-4.5	4	3	3.3	1.20	H	A1-1
    15	89,000	1	<6,000	3.3	0.50	V	A1-1

AMSU-B/MHS channel information::

    Channel / Centre Frequency (GHz) / Bandwidth (MHz) / NeDT (K) /
                                     Calibration Accuracy (K) / pol. angle (degree)
    16	89.0	<6000	1.0	1.0	90-q                  (Vertical pol)
    17	150	<4000	1.0	1.0	90-q                  (Vertical)
    18	183.31+-1.00	500	1.1	1.0	nospec        (Horizontal)
    19	183.31+-3.00	1000	1.0	1.0	nospec        (Horizontal)
    20	190.31+17.00	2000	1.2	1.0	90-q          (Vertical)

Since AMSU-A sensor is no longer available, we select only the AMSU-B/MHS
channels. We decide to use the same names of the five channels used for previous
NOAA MSPPS data files. i.e., select frequench index 15-19 (start from 0).

V1.0:  Initial version, NRL-MRY, January 26, 2021

Dataset information::

    Basic information on AMSU-B product file
    index: 1     2    3    4      5      6    7
    freq: 23.8,31.4,50.3,52.799,53.595,54.4,54.941,
    55.499,57.29,57.29,57.29,57.29,57.29,57.29,
    index:   8     9      10    11   12   13     14
    freq: 55.499,57.29,57.29,57.29,57.29,57.29,57.29,
    index: 15  16  17    18      19     20
    freq: 89.,89.,157.,183.311,183.311,190.311

    dimensions:
          Scanline = 2370 ;
          Field_of_view = 90 ;
          P_Layer = 100 ;
          Channel = 20 ;
          Qc_dim = 4 ;
    variables:
          Freq(Channel): Central Frequencies (GHz)
          Polo(Channel): Polarizations
          ScanTime_year(Scanline): Calendar Year 20XX
          ScanTime_doy(Scanline: julian day 1-366
          ScanTime_month(Scanline): Calendar month 1-12
          ScanTime_dom(Scanline): Calendar day of the month 1-31
          ScanTime_hour(Scanline: hour of the day 0-23
          ScanTime_minute(Scanline): minute of the hour 0-59
          ScanTime_second(Scanline): second of the minute 0-59
          ScanTime_UTC(Scanline): Number of seconds since 00:00:00 UTC
          Orb_mode(Scanline): 0-ascending,1-descending
          Latitude(Scanline, Field_of_view):Latitude of the view (-90,90),
                                            unit: degree
          Longitude(Scanline, Field_of_view):Longitude of the view (-180,180),
                                             unit: degree
          Sfc_type(Scanline, Field_of_view):type of surface:0-ocean,1-sea ice,
                                            2-land,3-snow
          Atm_type(Scanline, Field_of_view): type of atmosphere:currently missing
                                          ( note: not needed for geoips products)
          Qc(Scanline, Field_of_view, Qc_dim): Qc: 0-good, 1-usable with problem,
                                                   2-bad
          ChiSqr(Scanline, Field_of_view): Convergence rate: <3-good,>10-bad
          LZ_angle(Scanline, Field_of_view): Local Zenith Angle degree
          RAzi_angle(Scanline, Field_of_view):Relative Azimuth Angle 0-360 degree
          SZ_angle(Scanline, Field_of_view):Solar Zenith Angle (-90,90) degree
          BT(Scanline, Field_of_view, Channel): Channel Temperature (K)
          YM(Scanline, Field_of_view, Channel): UnCorrected Channel Temperature(K)
          ChanSel(Scanline, Field_of_view, Channel):Channels Selection Used in
                                                    Retrieval
          TPW(Scanline, Field_of_view): Total Precipitable Water (mm)
          CLW(Scanline, Field_of_view):Cloud liquid Water (mm)
          RWP(Scanline, Field_of_view): Rain Water Path (mm)
          LWP(Scanline, Field_of_view): Liquid Water Path (mm)
          SWP(Scanline, Field_of_view): Snow Water Path (mm)
          IWP(Scanline, Field_of_view): Ice Water Path (mm)
          GWP(Scanline, Field_of_view): Graupel Water Path (mm)
          RR(Scanline, Field_of_view): Rain Rate (mm/hr)
          Snow(Scanline, Field_of_view): Snow Cover (range: 0-1) i.e., 1 -> 100%
          SWE(Scanline, Field_of_view): Snow Water Equivalent (cm)
          SnowGS(Scanline, Field_of_view):Snow Grain Size (mm)
          SIce(Scanline, Field_of_view):Sea Ice Concentration (%)
          SIce_MY(Scanline, Field_of_view): Multi-Year Sea Ice Concentration (%)
          SIce_FY(Scanline, Field_of_view): First-Year Sea Ice Concentration (%)
          TSkin(Scanline, Field_of_view): Skin Temperature (K)
          SurfP(Scanline, Field_of_view): Surface Pressure (mb)
          Emis(Scanline, Field_of_view, Channel):Channel Emissivity
                                          (unit:1,  Emis:scale_factor = 0.0001)
          SFR(Scanline, Field_of_view): Snow Fall Rate in mm/hr
          CldTop(Scanline, Field_of_view): Cloud Top Pressure (scale_factor = 0.1)
          CldBase(Scanline, Field_of_view): Cloud Base Pressure
                                            (scale_factor = 0.1)
          CldThick(Scanline, Field_of_view): Cloud Thickness (scale_factor = 0.1)
          PrecipType(Scanline, Field_of_view): Precipitation Type (Frozen/Liquid)
          RFlag(Scanline, Field_of_view): Rain Flag
          SurfM(Scanline, Field_of_view): Surface Moisture (scale_factor = 0.1)
          WindSp(Scanline, Field_of_view): Wind Speed (m/s) (scale_factor = 0.01
          WindDir(Scanline, Field_of_view: Wind Direction (scale_factor = 0.01)
          WindU(Scanline, Field_of_view):U-direction Wind Speed (m/s)
                                        (scale_factor = 0.01)
          WindV(Scanline, Field_of_view: V-direction Wind Speed (m/s)
                                        (scale_factor = 0.01)
          Prob_SF(Scanline, Field_of_view): Probability of falling snow (%)

Additional info::

    Variables (nscan, npix):  npix=90 pixels per scan; nscan: vary with orbit
    chan-1 AT:  89 GHz              as ch16    anttenna temperature at V-pol
                                               FOV 16km at nadir
    chan-2 AT:  150 (157) GHz       as ch17 the number in bracket is for MHS
                                           from metops at V-pol, 16km at nadir
    chan-3 AT:  183.31 +-1 GHz          as ch18    at H-pol, 16km
    chan-4 AT:  183.31 +-3 GHz          as ch19    at H-pol, 16km
    chan-5 AT:  183.31 +-7 (190.3) GHz  as ch20    at V-pol, 16km
    lat:     -90, 90     deg
    lon:     -180, 180   deg
    RR:  surface rainrate (mm/hr)
    Snow: surafce snow cover
    IWP:  ice water path
    SWE:  snow water equvelent
    Sfc_type:  surface type
    Orbit_mode:   -1: ascending, 1: decending, 2: both
    SFR:  snowfall rate (unit?)
    LZ_angle:  local zinath angle (deg)
    SZ_angle:  solar zinath angle (deg)
"""
# Python Standard Libraries
import logging

# library for hdf files
import h5py

import matplotlib

matplotlib.use("agg")

LOG = logging.getLogger(__name__)

# Selected variables for TC products (add more if needed)
VARLIST = [
    "BT",
    "Latitude",
    "Longitude",
    "Freq",
    "RR",
    "TPW",
    "CLW",
    "RWP",
    "LWP",
    "SWP",
    "IWP",
    "GWP",
    "Snow",
    "SWE",
    "SIce",
    "SIce_MY",
    "SIce_FY",
    "TSkin",
    "SurfP",
    "Emis",
    "Sfc_type",
    "SFR",
    "CldTop",
    "CldBase",
    "CldThick",
    "WindSp",
    "WindDir",
    "WindU",
    "WindV",
    "LZ_angle",
    "RAzi_angle",
    "SZ_angle",
    "ScanTime_UTC",
    "ScanTime_second",
    "Orb_mode",
    "ScanTime_year",
    "ScanTime_doy",
    "ScanTime_month",
    "ScanTime_dom",
    "ScanTime_hour",
    "ScanTime_minute",
]

interface = "readers"
family = "standard"
name = "amsub_mirs"


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read AMSU/MHS MIRS data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    import os
    from datetime import datetime
    import numpy as np
    import pandas as pd
    import xarray as xr

    xarrays = []
    for fname in fnames:
        LOG.info("Reading file %s", fname)

        #     check for right input AMSU-B/MHS data file
        data_name = os.path.basename(fname).split("_")[-1].split(".")[-1]

        if data_name != "nc":
            LOG.info("Warning: wrong AMSU-B/MHS data type:  data_type=", data_name)
            raise

        if "NPR-MIRS-IMG" in os.path.basename(fname):
            LOG.info("found a NOAA MIRS AMSU-B/MHS file")
        else:
            LOG.info("not a NOAA MIRS AMSU-B/MHS file: skip it")
            raise

        """    ------  Notes  ------
        Read AMSU-B hdf files for 5 chan antenna temperature (AT) and asscoaited EDRs
            Then, transform these ATs and fields into xarray framework for GEOIPS
            ( AT will be corrected into brightness temperature (TB) later)

        Input Parameters:
            fname (str): input file name.  require a full path
        Returns:
            xarray.Dataset with required Variables and Attributes:
                Variables:
                            AMSUB vars:
                            'latitude', 'longitude', 'Ch1', 'Ch2', 'Ch3', 'Ch4','Ch4',
                            'RR', 'Snow','SWE','IWP','SFR' 'sfcType', 'time_scan'
                Attibutes:
                            'source_name', 'platform_name', 'data_provider',
                            'interpolation_radius_of_influence', 'start_datetime',
                            'end_datetime'
        """

        # import all data contents
        fileobj = h5py.File(fname, mode="r")

        var_all = {}

        # read-in selected variables
        for var in VARLIST:
            var_all[var] = fileobj[var]

        # values adjustment with scale_factor if there is a "scale_factor'

        for var in VARLIST:
            var_attrs = var_all[var].attrs.items()  # get attribute info
            for attrs in var_attrs:  # loop attribute tuples
                if "scale_factor" in attrs[0]:
                    factor = attrs[1][0]
                    var_all[var] = var_all[var] * factor

        # Select 5 AMSU-B/MHS Channels

        Chan1_AT = var_all["BT"][:, :, 15]
        Chan2_AT = var_all["BT"][:, :, 16]
        Chan3_AT = var_all["BT"][:, :, 17]
        Chan4_AT = var_all["BT"][:, :, 18]
        Chan5_AT = var_all["BT"][:, :, 19]

        # start_time=os.path.basename(fname).split('_')[3][1:13]
        # end_time=os.path.basename(fname).split('_')[4][1:13]
        # Collect start_time and end_time of input file (one orbit) from
        # attributes - filename wrong!
        start_datetime = datetime.strptime(
            fileobj.attrs["time_coverage_start"].astype(str), "%Y-%m-%dT%H:%M:%SZ"
        )
        end_datetime = datetime.strptime(
            fileobj.attrs["time_coverage_end"].astype(str), "%Y-%m-%dT%H:%M:%SZ"
        )

        #  -------- Apply the GEOIPS framework in XARRAY data frame ----------

        LOG.info("Making full dataframe")

        # setup the time in datetime64 format
        npix = var_all["RR"].shape[1]  # pixels per scan
        nscan = var_all["RR"].shape[0]  # total scans of this file

        time_scan = np.zeros((nscan, npix))
        # take time of each scan
        for i in range(nscan):
            yr = var_all["ScanTime_year"][i]
            dy = var_all["ScanTime_doy"][i]
            hr = var_all["ScanTime_hour"][i]
            mn = var_all["ScanTime_minute"][i]
            try:
                time_scan[i:] = "%04d%03d%02d%02d" % (yr, dy, hr, mn)
            except ValueError:
                LOG.info(
                    "Could not parse time for scan line %s: YEAR=%s, DOY=%s, "
                    "HOUR=%s, MINUTE=%s",
                    i,
                    yr,
                    dy,
                    hr,
                    mn,
                )
                continue
        #          ------  setup xarray variables   ------

        # namelist_amsub  = ['latitude', 'longitude', 'Chan1_AT', 'Chan2_AT', 'Chan3_AT'
        #                    ,'Chan4_AT','Chan5_AT', 'RR','Snow','IWP','SWE','SFR',
        #                    'Sfc_type','time']

        xarray_amsub = xr.Dataset()

        # setup attributes

        # satID and start_end time from input filename (ma3 --> METOP-c?) - TBD:
        # depending on data file )
        sat_id = os.path.basename(fname).split("_")[2]
        if sat_id == "n19":
            satid = "noaa-19"
            crtm_name = "mhs_n19"
        elif sat_id == "n18":
            satid = "noaa-18"
            crtm_name = "mhs_n18"
        elif sat_id == "n20":
            satid = "noaa-20"
            crtm_name = "mhs_n20"
        elif sat_id == "ma1":
            satid = "metop-b"
            crtm_name = "mhs_metob-b"
        elif sat_id == "ma2":
            satid = "metop-a"
            crtm_name = "mhs_metop-a"
        elif sat_id == "ma3":
            satid = "metop-c"
            crtm_name = "mhs_metop-c"

        # add attributes to xarray
        xarray_amsub.attrs["start_datetime"] = start_datetime
        xarray_amsub.attrs["end_datetime"] = end_datetime
        xarray_amsub.attrs["source_name"] = "amsu-b"
        xarray_amsub.attrs["platform_name"] = satid
        xarray_amsub.attrs["crtm_name"] = crtm_name
        xarray_amsub.attrs["data_provider"] = "NOAA-nesdis"

        # MTIFs need to be "prettier" for PMW products, so 2km resolution for
        # final image.
        # xarray_amsub.attrs['sample_distance_km'] = 15
        xarray_amsub.attrs["sample_distance_km"] = 2
        xarray_amsub.attrs["interpolation_radius_of_influence"] = 30000

        if metadata_only:
            LOG.info("metadata_only requested, returning without reading data")
            return {"METADATA": xarray_amsub}

        # keep same variables from previous version of amsub reader for files from MSPPS
        xarray_amsub["latitude"] = xr.DataArray(var_all["Latitude"][()])
        xarray_amsub["longitude"] = xr.DataArray(var_all["Longitude"][()])
        xarray_amsub["Chan1_AT"] = xr.DataArray(Chan1_AT, attrs={"channel_number": 1})
        xarray_amsub["Chan2_AT"] = xr.DataArray(Chan2_AT, attrs={"channel_number": 2})
        xarray_amsub["Chan3_AT"] = xr.DataArray(Chan3_AT, attrs={"channel_number": 3})
        xarray_amsub["Chan4_AT"] = xr.DataArray(Chan4_AT, attrs={"channel_number": 4})
        xarray_amsub["Chan5_AT"] = xr.DataArray(Chan5_AT, attrs={"channel_number": 5})
        xarray_amsub["RR"] = xr.DataArray(var_all["RR"][()])
        xarray_amsub["Snow"] = xr.DataArray(var_all["Snow"][()])
        xarray_amsub["IWP"] = xr.DataArray(var_all["IWP"][()])
        xarray_amsub["SWE"] = xr.DataArray(var_all["SWE"][()])
        xarray_amsub["SFR"] = xr.DataArray(var_all["SFR"][()])
        xarray_amsub["sfcType"] = xr.DataArray(var_all["Sfc_type"][()])
        xarray_amsub["time"] = xr.DataArray(
            pd.DataFrame(time_scan).astype(int).apply(pd.to_datetime, format="%Y%j%H%M")
        )

        # add variables from MIRS file
        xarray_amsub["TPW"] = xr.DataArray(var_all["TPW"][()])
        xarray_amsub["CLW"] = xr.DataArray(var_all["CLW"][()])
        xarray_amsub["RWP"] = xr.DataArray(var_all["RWP"][()])
        xarray_amsub["GWP"] = xr.DataArray(var_all["GWP"][()])
        xarray_amsub["SIce"] = xr.DataArray(var_all["SIce"][()])
        xarray_amsub["SIce_MY"] = xr.DataArray(var_all["SIce_MY"][()])
        xarray_amsub["SIce_FY"] = xr.DataArray(var_all["SIce_FY"][()])
        xarray_amsub["TSkin"] = xr.DataArray(var_all["TSkin"][()])
        xarray_amsub["SurfP"] = xr.DataArray(var_all["SurfP"][()])
        xarray_amsub["CldTop"] = xr.DataArray(var_all["CldTop"][()])
        xarray_amsub["CldBase"] = xr.DataArray(var_all["CldBase"][()])
        xarray_amsub["CldThick"] = xr.DataArray(var_all["CldThick"][()])
        xarray_amsub["WindSp"] = xr.DataArray(var_all["WindSp"][()])
        xarray_amsub["WindDir"] = xr.DataArray(var_all["WindDir"][()])
        xarray_amsub["WindU"] = xr.DataArray(var_all["WindU"][()])
        xarray_amsub["WindV"] = xr.DataArray(var_all["WindV"][()])
        xarray_amsub["satellite_zenith_angle"] = xr.DataArray(var_all["LZ_angle"][()])
        xarray_amsub["SZ_angle"] = xr.DataArray(var_all["SZ_angle"][()])
        xarray_amsub["RAzi_angle"] = xr.DataArray(var_all["RAzi_angle"][()])
        # from amsub_mhs_prep/oned_innov.f90:
        beam_pos = np.broadcast_to(
            np.arange(fileobj["Field_of_view"].size) + 1, var_all["LZ_angle"][()].shape
        )
        xarray_amsub["sensor_scan_angle"] = xr.DataArray((beam_pos - 45.5) * 10.0 / 9.0)

        xarrays.append(xarray_amsub)

    final_xarray = xr.concat(xarrays, dim="dim_0")

    return {"AMSUB": final_xarray, "METADATA": final_xarray[[]]}



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Readers interface module."""

from geoips.interfaces.base import BaseModuleInterface


class ReadersInterface(BaseModuleInterface):
    """Interface for ingesting a specific data type.

    Provides specification for ingensting a specific data type, and storing in
    the GeoIPS xarray-based internal format.
    """

    name = "readers"
    required_args = {"standard": ["fnames"]}
    required_kwargs = {
        "standard": ["metadata_only", "chans", "area_def", "self_register"]
    }


readers = ReadersInterface()



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module containing colormap for 89pct product."""

import logging

LOG = logging.getLogger(__name__)

interface = "colormappers"
family = "matplotlib"
name = "pmw_89pct"


def call(data_range=[105, 280], cbar_label="TB (K)"):
    """Colormap for displaying ~89GHz PMW data for weak TCs.

    Parameters
    ----------
    data_range : list of float, default=[105, 280]
        * Min and max value for colormap.
        * Ensure the data range matches the range of the algorithm specified for
          use with this colormap
        * This colormap MUST include 105 and 280

    Returns
    -------
    mpl_colors_info : dict
        Dictionary of matplotlib plotting parameters, to ensure consistent
        image output
    """
    min_tb = data_range[0]
    max_tb = data_range[1]

    if min_tb >= 125 or max_tb <= 265:
        raise ("89pct range must include 125 and 265")

    from geoips.image_utils.colormap_utils import create_linear_segmented_colormap

    transition_vals = [
        (min_tb, 125),
        (125, 150),
        (150, 175),
        (175, 212),
        (212, 230),
        (230, 250),
        (250, 265),
        (265, max_tb),
    ]
    # (280, max_tb)]
    transition_colors = [
        ("orange", "chocolate"),
        ("chocolate", "indianred"),
        ("indianred", "firebrick"),
        ("firebrick", "red"),
        ("gold", "yellow"),
        ("lime", "limegreen"),
        ("deepskyblue", "blue"),
        ("navy", "slateblue"),
    ]
    # ('magenta', 'white')]

    ticks = [int(xx[0]) for xx in transition_vals]

    # special selection of label

    # ticks = [min_tb, 125, 150, 175, 200, 225, 250, 275, max_tb]

    # selection of min and max values for colormap if needed
    min_tb = transition_vals[0][0]
    max_tb = transition_vals[-1][1]
    ticks = ticks + [int(max_tb)]

    LOG.info("Setting cmap")
    mpl_cmap = create_linear_segmented_colormap(
        "89pct_cmap", min_tb, max_tb, transition_vals, transition_colors
    )

    LOG.info("Setting norm")
    from matplotlib.colors import Normalize

    mpl_norm = Normalize(vmin=min_tb, vmax=max_tb)

    # Must be uniform or proportional, None not valid for Python 3
    cbar_spacing = "proportional"
    mpl_tick_labels = None
    mpl_boundaries = None

    # from geoips.image_utils.mpl_utils import create_colorbar
    # only create colorbar for final imagery
    # cbar = create_colorbar(fig, mpl_cmap, mpl_norm, ticks, cbar_label=cbar_label)
    mpl_colors_info = {
        "cmap": mpl_cmap,
        "norm": mpl_norm,
        "cbar_ticks": ticks,
        "cbar_tick_labels": mpl_tick_labels,
        "cbar_label": cbar_label,
        "boundaries": mpl_boundaries,
        "cbar_spacing": cbar_spacing,
        "colorbar": True,
        "cbar_full_width": True,
    }

    # return cbar, min_tb, max_tb
    return mpl_colors_info



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""PluginRegistry class to interface with the JSON plugin registries.

The "create_plugin_registries" utility generates a JSON file at the top
level of every geoips plugin package with a complete list of all plugins with
the associated metadata (everything except the actual contents of the plugin
itself).

Once all of the registered_plugins.json files have been generated via
create_plugin_registries, this class uses those registries to quickly
identify and open plugins as required.  Previously the individual
interface classes would open all plugins every time one was required,
so moving this process into a single PluginRegistry object allows us to
more effectively cache plugins across all interfaces, and avoid reading
in all plugins multiple times.
"""

from importlib import metadata, resources
import os
import logging
from geoips.errors import PluginRegistryError
import yaml
import json


LOG = logging.getLogger(__name__)


class PluginRegistry:
    """Plugin Registry class definition.

    Represents all of the plugins found in all of the available GeoIPS packages.
    This class will load a plugin when requested, rather than loading all plugins when
    GeoIPS is instantiated.
    """

    def __init__(self, _test_registry_files=[]):
        # Use this for unit testing
        if _test_registry_files:
            self.registry_files = _test_registry_files
            self._is_test = True
        # Use this for normal operation and collect the registry files
        else:
            self._is_test = False
            self.registry_files = []  # Collect the paths to the registry files here
            for pkg in metadata.entry_points(group="geoips.plugin_packages"):
                try:
                    self.registry_files.append(
                        str(resources.files(pkg.value) / "registered_plugins.json")
                    )
                except TypeError:
                    raise PluginRegistryError(
                        f"resources.files('{pkg.value}') failed\n"
                        f"pkg {pkg}\n"
                        "Potentially missing __init__.py file? Try:\n"
                        f"    touch {pkg.value}/__init__.py\n"
                        "and try again.\n"
                        "Note you will need to add a docstring to "
                        f"{pkg.value}/__init__.py in order for all tests to pass"
                    )

    @property
    def registered_plugins(self):
        """Plugin Registry registered_plugins attribute."""
        if not hasattr(self, "_registered_plugins"):
            self._set_class_properties()
        return self._registered_plugins

    @property
    def interface_mapping(self):
        """Plugin Registry interface_mapping attribute."""
        if not hasattr(self, "_interface_mapping"):
            self._set_class_properties()
        return self._interface_mapping

    def _set_class_properties(self):
        """Find all plugins in registered plugin packages.

        Search the ``registered_plugins.yaml`` of each registered plugin package and add
        them to the _registered_plugins dictionary
        """
        # Load the registries here and return them as a dictionary
        if not hasattr(self, "_registered_plugins"):
            from geoips.geoips_utils import merge_nested_dicts

            # Complete dictionary of all available plugins found in every geoips package
            self._registered_plugins = {}
            # A mapping of interfaces to plugin_types. Ie:
            # {
            # "yaml_based": [products, sectors, ...],
            # "module_based": [algorithms, readers, ...],
            # "text_based": [tpw_cimss, ...]
            # }
            self._interface_mapping = {}
            for reg_path in self.registry_files:
                if not os.path.exists(reg_path):
                    LOG.error(
                        f"Plugin registry {reg_path} did not exist, "
                        "please run 'create_plugin_registries'"
                    )
                    continue
                # This will include all plugins, including schemas, yaml_based,
                # and module_based plugins.
                if self._is_test:
                    pkg_plugins = yaml.safe_load(open(reg_path, "r"))
                else:
                    pkg_plugins = json.load(open(reg_path, "r"))
                    # Do not validate ALL plugins at runtime.
                    # self.validate_registry(pkg_plugins, reg_path)
                try:
                    for plugin_type in pkg_plugins:
                        if plugin_type not in self._registered_plugins:
                            self._registered_plugins[plugin_type] = {}
                            self._interface_mapping[plugin_type] = []
                        for interface in pkg_plugins[plugin_type]:
                            interface_dict = pkg_plugins[plugin_type][interface]
                            if interface not in self._registered_plugins[plugin_type]:
                                self._registered_plugins[plugin_type][
                                    interface
                                ] = interface_dict  # NOQA
                                self._interface_mapping[plugin_type].append(interface)
                            else:
                                merge_nested_dicts(
                                    self._registered_plugins[plugin_type][interface],
                                    interface_dict,
                                )
                except TypeError:
                    raise PluginRegistryError(f"Failed reading {reg_path}.")
            # Let's test this separately, not at runtime (see validate_all_registries).
            # Assume it was tested up front, and no longer needs testing at
            # runtime, so we don't fail catastrophically for a single bad
            # plugin that may not even be used.
            # if not self._is_test:
            #     self.validate_registry(
            #         self._registered_plugins,
            #         "all_registered_plugins",
            #     )
        return self._registered_plugins

    def get_plugin_info(self, interface, plugin_name):
        """Find a plugin in the registry and return its info.

        This should remove all plugin loading from the base interfaces and allow us
        to only load one plugin at a time
        """
        plugin_type = self.identify_plugin_type(interface)
        if isinstance(plugin_name, tuple):
            return self.registered_plugins[plugin_type][interface][plugin_name[0]][
                plugin_name[1]
            ]
        return self.registered_plugins[plugin_type][interface][plugin_name]

    def get_interface_plugin_info(self, interface):
        """Find an interface in the registry and return its corresponding info.

        This should remove all plugin loading from the base interfaces and allow us
        to only load one plugin at a time
        """
        plugin_type = self.identify_plugin_type(interface)
        return self.registered_plugins[plugin_type][interface]

    def get_interface_plugin_names(self, interface):
        """List all available plugins in the an interface of the registry.

        This should remove all plugin loading from the base interfaces and allow us
        to only load one plugin at a time
        """
        plugin_type = self.identify_plugin_type(interface)
        return list(self.registered_plugins[plugin_type][interface].keys())

    def list_plugins(self, interface):
        """List the plugins available for an interface ONLY based on the registries.

        This should not load any plugins, just return info from the registries.
        """
        plugin_type = self.identify_plugin_type(interface)
        return self.registered_plugins[plugin_type][interface]

    def identify_plugin_type(self, interface):
        """Identify the Plugin Type based on the provided interface."""
        plugin_type = None
        for p_type in self.interface_mapping:
            if interface in self.interface_mapping[p_type]:
                plugin_type = p_type
                break
        if plugin_type is None:
            raise PluginRegistryError(
                f"{interface} does not exist within any package registry."
            )
        return plugin_type


plugin_registry = PluginRegistry()
